"""Unit tests for djblets.util.json_utils."""

from __future__ import unicode_literals

from django.utils import six

from djblets.testing.testcases import TestCase
from djblets.util.json_utils import (JSONPatchError,
                                     JSONPatchPathError,
                                     JSONPatchReadAccessError,
                                     JSONPatchTestError,
                                     JSONPatchWriteAccessError,
                                     JSONPointerEndOfList,
                                     JSONPointerLookupError,
                                     JSONPointerSyntaxError,
                                     json_get_pointer_info,
                                     json_merge_patch,
                                     json_patch,
                                     json_resolve_pointer)


class JSONMergePatchTests(TestCase):
    """Unit tests for djblets.util.json_utils.json_merge_patch."""

    def test_with_change_dict_key_value(self):
        """Testing json_merge_patch with changing key's value in dictionary"""
        doc = {
            'a': 'b',
            'b': ['c'],
            'c': 'd',
            'd': [1, 2],
            'e': {
                'f': 1,
                'g': 2,
            },
        }

        self.assertEqual(
            json_merge_patch(
                doc=doc,
                patch={
                    'a': 'c',
                    'b': 'd',
                    'c': ['e'],
                    'd': [3, 4],
                    'e': {
                        'f': 'ef',
                    },
                }
            ),
            {
                'a': 'c',
                'b': 'd',
                'c': ['e'],
                'd': [3, 4],
                'e': {
                    'f': 'ef',
                    'g': 2,
                },
            })

        self.assertEqual(
            doc,
            {
                'a': 'b',
                'b': ['c'],
                'c': 'd',
                'd': [1, 2],
                'e': {
                    'f': 1,
                    'g': 2,
                },
            })

    def test_with_add_dict_key(self):
        """Testing json_merge_patch with adding a new key to a dictionary"""
        doc = {
            'a': 'b',
        }

        self.assertEqual(
            json_merge_patch(
                doc=doc,
                patch={
                    'b': 'c',
                    'c': ['d'],
                    'd': {
                        'e': 1,
                        'f': 2,
                    },
                }
            ),
            {
                'a': 'b',
                'b': 'c',
                'c': ['d'],
                'd': {
                    'e': 1,
                    'f': 2,
                },
            })

        self.assertEqual(
            doc,
            {
                'a': 'b',
            })

    def test_with_delete_dict_key(self):
        """Testing json_merge_patch with deleting a key in a dictionary"""
        doc = {
            'a': 'b',
            'b': 'c',
            'c': {
                'e': 1,
                'f': 2,
            },
            'd': None,
        }

        self.assertEqual(
            json_merge_patch(
                doc=doc,
                patch={
                    'a': None,
                    'c': {
                        'e': None,
                    },
                    'e': None,
                }
            ),
            {
                'b': 'c',
                'c': {
                    'f': 2,
                },
                'd': None,
            })

        self.assertEqual(
            doc,
            {
                'a': 'b',
                'b': 'c',
                'c': {
                    'e': 1,
                    'f': 2,
                },
                'd': None,
            })

    def test_with_change_list(self):
        """Testing json_merge_patch with changing a list value"""
        doc = [1, 2, 3]

        self.assertEqual(
            json_merge_patch(
                doc=doc,
                patch=[4, 5, 6],
            ),
            [4, 5, 6])

        self.assertEqual(doc, [1, 2, 3])

    def test_with_change_type(self):
        """Testing json_merge_patch with changing a value's type"""
        doc = [1, 2, 3]

        self.assertEqual(
            json_merge_patch(
                doc=doc,
                patch={
                    'a': 1,
                }
            ),
            {
                'a': 1,
            })

        self.assertEqual(doc, [1, 2, 3])

    def test_with_can_write_key_func(self):
        """Testing json_merge_patch with can_write_key_func"""
        def _can_write_key(doc, patch, path):
            return path != ('a', 'b', 'c')

        doc = {
            'a': {
                'b': {
                    'c': 123,
                },
                'd': True,
            },
        }

        self.assertEqual(
            json_merge_patch(
                doc=doc,
                patch={
                    'a': {
                        'b': {
                            'c': 100,
                        },
                        'd': False,
                    },
                    'e': 'hi!',
                },
                can_write_key_func=_can_write_key),
            {
                'a': {
                    'b': {
                        'c': 123,
                    },
                    'd': False,
                },
                'e': 'hi!',
            })

        self.assertEqual(
            doc,
            {
                'a': {
                    'b': {
                        'c': 123,
                    },
                    'd': True,
                },
            })

    def test_with_can_write_key_func_not_callable(self):
        """Testing json_merge_patch with can_write_key_func not callable"""
        message = 'can_write_key_func must be callable'

        with self.assertRaisesMessage(ValueError, message):
            json_merge_patch(doc={},
                             patch={},
                             can_write_key_func=123)

    def test_with_can_write_key_func_with_patch_error(self):
        """Testing json_merge_patch with can_write_key_func raising
        JSONPatchError
        """
        def _can_write_key(doc, patch, path):
            if path == ('a', 'b', 'c'):
                raise JSONPatchError('Go away!',
                                     doc=doc,
                                     patch=patch)

            return True

        doc = {
            'a': {
                'b': {
                    'c': 123,
                },
                'd': True,
            },
        }

        with self.assertRaisesMessage(JSONPatchError, 'Go away!'):
            json_merge_patch(
                doc=doc,
                patch={
                    'a': {
                        'b': {
                            'c': 100,
                        },
                        'd': False,
                    },
                    'e': 'hi!',
                },
                can_write_key_func=_can_write_key)

        self.assertEqual(doc, {
            'a': {
                'b': {
                    'c': 123,
                },
                'd': True,
            },
        })


class JSONPatchTests(TestCase):
    """Unit tests for djblets.util.json_utils.json_patch."""

    def test_with_patch_not_list(self):
        """Testing json_patch with patch not a list"""
        message = 'The patch must be a list of operations to perform'

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc={}, patch={})

        e = cm.exception
        self.assertEqual(e.doc, {})
        self.assertEqual(e.patch, {})
        self.assertIsNone(e.patch_entry_index)

    def test_with_patch_entry_not_dict(self):
        """Testing json_patch with patch entry not a dictionary"""
        message = 'Patch entry 0 must be a dictionary instead of list'

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc={}, patch=[[]])

        e = cm.exception
        self.assertEqual(e.doc, {})
        self.assertEqual(e.patch, [[]])
        self.assertEqual(e.patch_entry_index, 0)

    def test_with_patch_entry_missing_op(self):
        """Testing json_patch with patch entry missing op"""
        message = 'Missing key "op" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc={}, patch=[{}])

        e = cm.exception
        self.assertEqual(e.doc, {})
        self.assertEqual(e.patch, [{}])
        self.assertEqual(e.patch_entry_index, 0)

    def test_with_dict_and_empty_patch(self):
        """Testing json_patch with dictionary object and empty patch"""
        doc = {
            'foo': 1,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[]),
            {
                'foo': 1,
            })
        self.assertEqual(
            doc,
            {
                'foo': 1,
            })

    def test_with_list_and_empty_patch(self):
        """Testing json_patch with list object and empty patch"""
        doc = ['foo']

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[]),
            ['foo'])

    def test_add_replacing_field(self):
        """Testing json_patch with add op replacing existing field"""
        doc = {
            'foo': None,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'add',
                    'path': '/foo',
                    'value': 1,
                }]),
            {
                'foo': 1,
            })
        self.assertEqual(
            doc,
            {
                'foo': None,
            })

    def test_add_to_top_level_dict(self):
        """Testing json_patch with add op in top-level dictionary"""
        doc = {}

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'add',
                    'path': '/foo',
                    'value': 'bar',
                }]),
            {
                'foo': 'bar',
            })
        self.assertEqual(doc, {})

    def test_add_to_top_level_array(self):
        """Testing json_patch with add op in top-level array"""
        doc = []

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'add',
                    'path': '/0',
                    'value': 'foo',
                }]),
            ['foo'])
        self.assertEqual(doc, [])

    def test_add_with_append_to_top_level_array(self):
        """Testing json_patch with add op with appending to top-level array"""
        doc = ['a', 'b']

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'add',
                    'path': '/-',
                    'value': 'c',
                }]),
            ['a', 'b', 'c'])
        self.assertEqual(doc, ['a', 'b'])

    def test_add_with_insert_to_top_level_array(self):
        """Testing json_patch with add op with inserting into top-level array
        """
        doc = ['a', 'b']

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'add',
                    'path': '/1',
                    'value': 'c',
                }]),
            ['a', 'c', 'b'])
        self.assertEqual(doc, ['a', 'b'])

    def test_add_with_empty_key_on_top_level(self):
        """Testing json_patch with add op with top level and "/" path
        (signifying empty key)
        """
        doc = {}

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'add',
                    'path': '/',
                    'value': 1,
                }]),
            {
                '': 1,
            })
        self.assertEqual(doc, {})

    def test_add_with_empty_key_on_child(self):
        """Testing json_patch with add op with child dictionary and "/" path
        (signifying empty key)
        """
        doc = {
            'foo': {},
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'add',
                    'path': '/foo/',
                    'value': 1,
                }]),
            {
                'foo': {
                    '': 1,
                },
            })
        self.assertEqual(doc, {
            'foo': {},
        })

    def test_add_with_complex_path(self):
        """Testing json_patch with add op with complex path"""
        doc = {
            'a': 1,
            'b': [{
                'c': 2,
            }],
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'add',
                    'path': '/b/0/d',
                    'value': 100,
                }]),
            {
                'a': 1,
                'b': [{
                    'c': 2,
                    'd': 100,
                }],
            })
        self.assertEqual(doc, {
            'a': 1,
            'b': [{
                'c': 2,
            }],
        })

    def test_add_with_dict_and_numeric_key(self):
        """Testing json_patch with add op with dictionary and numeric key"""
        doc = {
            'foo': 1,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'add',
                    'path': '/0',
                    'value': 2,
                }]),
            {
                'foo': 1,
                '0': 2,
            })
        self.assertEqual(doc, {
            'foo': 1,
        })

    def test_add_with_array_and_index_0(self):
        """Testing json_patch with add op with array and inserting at index 0
        """
        doc = {
            'a': [1, 2, 3],
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'add',
                    'path': '/a/0',
                    'value': 100,
                }]),
            {
                'a': [100, 1, 2, 3],
            })
        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_add_with_array_and_last_index_plus_1(self):
        """Testing json_patch with add op with array and inserting at last
        index + 1
        """
        doc = {
            'a': [1, 2, 3],
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'add',
                    'path': '/a/3',
                    'value': 100,
                }]),
            {
                'a': [1, 2, 3, 100],
            })
        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_add_with_can_write_key_func_and_allowed(self):
        """Testing json_patch with add op and can_write_key_func and key
        allowed
        """
        def _can_write_key(doc, patch_entry, path):
            return path == ('a', 'c')

        doc = {
            'a': {
                'b': 1,
            },
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'add',
                    'path': '/a/c',
                    'value': 2,
                }],
                can_write_key_func=_can_write_key),
            {
                'a': {
                    'b': 1,
                    'c': 2,
                },
            })

    def test_add_with_can_write_key_func_and_not_allowed(self):
        """Testing json_patch with add op and can_write_key_func and key
        not allowed
        """
        def _can_write_key(doc, patch_entry, path):
            return path == ('a', 'b')

        doc = {
            'a': {
                'b': 1,
            },
        }
        patch = [{
            'op': 'add',
            'path': '/a/c',
            'value': 2,
        }]

        message = 'Cannot write to path "/a/c" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchWriteAccessError,
                                      message) as cm:
            json_patch(doc=doc,
                       patch=patch,
                       can_write_key_func=_can_write_key)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/c')

    def test_add_with_array_and_index_out_of_bounds(self):
        """Testing json_patch with add op with array and index out of bounds"""
        doc = {
            'a': [1, 2, 3],
        }
        patch = [{
            'op': 'add',
            'path': '/a/4',
            'value': 100,
        }]

        message = 'Cannot insert into index 4 in path "/a/4" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/4')

        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_add_with_array_and_negative_index(self):
        """Testing json_patch with add op with array and negative index"""
        doc = {
            'a': [1, 2, 3],
        }
        patch = [{
            'op': 'add',
            'path': '/a/-1',
            'value': 100,
        }]

        message = (
            'Syntax error in path "/a/-1" for patch entry 0: Negative indexes '
            'into lists are not allowed'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/-1')

        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_add_with_array_and_key(self):
        """Testing json_patch with add op with array and attempting to insert
        at key
        """
        doc = {
            'a': [1, 2, 3],
        }
        patch = [{
            'op': 'add',
            'path': '/a/bad',
            'value': 100,
        }]

        message = (
            'Syntax error in path "/a/bad" for patch entry 0: %r is not a '
            'valid list index in "/a"'
            % 'bad'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/bad')

        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_add_with_missing_path(self):
        """Testing json_patch with add op and missing path"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'add',
            'value': 1,
        }]

        message = 'Missing key "path" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_add_with_missing_value(self):
        """Testing json_patch with add op and missing value"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'add',
            'path': '/a',
        }]

        message = 'Missing key "value" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_add_with_bad_path(self):
        """Testing json_patch with add op and bad path"""
        doc = {
            'a': 1,
        }

        patch = [{
            'op': 'add',
            'path': '/b/c',
            'value': 100,
        }]

        message = (
            'Invalid path "/b/c" for patch entry 0: Dictionary key "b" not '
            'found in "/"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/b/c')

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_add_with_invalid_path_syntax(self):
        """Testing json_patch with add op and invalid path syntax"""
        doc = {}
        patch = [{
            'op': 'add',
            'path': 'b',
            'value': 100,
        }]

        message = (
            'Syntax error in path "b" for patch entry 0: Paths must either '
            'be empty or start with a "/"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, 'b')

    def test_remove_with_dict_key(self):
        """Testing json_patch with remove op with dictionary key"""
        doc = {
            'a': 1,
            'b': 2,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'remove',
                    'path': '/a',
                }]),
            {
                'b': 2,
            })
        self.assertEqual(doc, {
            'a': 1,
            'b': 2,
        })

    def test_remove_with_array_index_0(self):
        """Testing json_patch with remove op with array index 0"""
        doc = {
            'a': [1, 2, 3],
            'b': 2,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'remove',
                    'path': '/a/0',
                }]),
            {
                'a': [2, 3],
                'b': 2,
            })
        self.assertEqual(doc, {
            'a': [1, 2, 3],
            'b': 2,
        })

    def test_remove_with_array_index_middle(self):
        """Testing json_patch with remove op with array index (middle of array)
        """
        doc = {
            'a': [1, 2, 3],
            'b': 2,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'remove',
                    'path': '/a/1',
                }]),
            {
                'a': [1, 3],
                'b': 2,
            })
        self.assertEqual(doc, {
            'a': [1, 2, 3],
            'b': 2,
        })

    def test_remove_with_array_index_last(self):
        """Testing json_patch with remove op with array index (last in array)
        """
        doc = {
            'a': [1, 2, 3],
            'b': 2,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'remove',
                    'path': '/a/2',
                }]),
            {
                'a': [1, 2],
                'b': 2,
            })
        self.assertEqual(doc, {
            'a': [1, 2, 3],
            'b': 2,
        })

    def test_remove_with_array_index_end(self):
        """Testing json_patch with remove op with array index `-` (end of
        array)
        """
        doc = {
            'a': [1, 2, 3],
            'b': 2,
        }
        patch = [{
            'op': 'remove',
            'path': '/a/-',
        }]

        message = (
            'Cannot perform operation "remove" on end of list at "/a/-" for '
            'patch entry 0'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/-')

        self.assertEqual(doc, {
            'a': [1, 2, 3],
            'b': 2,
        })

    def test_remove_with_complex_path(self):
        """Testing json_patch with remove op with complex path"""
        doc = {
            'a': [{
                'b': 1,
                'c': 2,
            }],
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'remove',
                    'path': '/a/0/b'
                }]),
            {
                'a': [{
                    'c': 2,
                }],
            })
        self.assertEqual(doc, {
            'a': [{
                'b': 1,
                'c': 2,
            }],
        })

    def test_remove_with_repeated_removes(self):
        """Testing json_patch with remove op multiple times with different
        paths
        """
        doc = [1, 2, 3, 4]

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[
                    {
                        'op': 'remove',
                        'path': '/1',
                    },
                    {
                        'op': 'remove',
                        'path': '/2',
                    },
                ]),
            [1, 3])
        self.assertEqual(doc, [1, 2, 3, 4])

    def test_remove_with_can_write_key_func_and_allowed(self):
        """Testing json_patch with remove op and can_write_key_func and key
        allowed
        """
        def _can_write_key(doc, patch_entry, path):
            return path == ('a', 'b')

        doc = {
            'a': {
                'b': 1,
            },
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'remove',
                    'path': '/a/b',
                }],
                can_write_key_func=_can_write_key),
            {
                'a': {},
            })

    def test_remove_with_can_write_key_func_and_not_allowed(self):
        """Testing json_patch with remove op and can_write_key_func and key
        not allowed
        """
        def _can_write_key(doc, patch_entry, path):
            return path == ('a', 'c')

        doc = {
            'a': {
                'b': 1,
            },
        }
        patch = [{
            'op': 'remove',
            'path': '/a/b',
        }]

        message = 'Cannot write to path "/a/b" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchWriteAccessError,
                                      message) as cm:
            json_patch(doc=doc,
                       patch=patch,
                       can_write_key_func=_can_write_key)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/b')

    def test_remove_with_missing_path(self):
        """Testing json_patch with remove op and missing path"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'remove',
            'value': 1,
        }]

        message = 'Missing key "path" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_remove_with_bad_path(self):
        """Testing json_patch with remove op and bad path"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'remove',
            'path': '/b',
        }]

        message = (
            'Invalid path "/b" for patch entry 0: Dictionary key "b" not '
            'found in "/"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/b')

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_remove_with_invalid_path_syntax(self):
        """Testing json_patch with remove op and invalid path syntax"""
        doc = {}
        patch = [{
            'op': 'remove',
            'path': 'b',
        }]

        message = (
            'Syntax error in path "b" for patch entry 0: Paths must either '
            'be empty or start with a "/"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, 'b')

    def test_remove_with_array_and_index_out_of_bounds(self):
        """Testing json_patch with remove op with array and index out of bounds
        """
        doc = {
            'a': [1, 2],
        }
        patch = [{
            'op': 'remove',
            'path': '/a/3',
        }]

        message = (
            'Invalid path "/a/3" for patch entry 0: 3 is outside the list in '
            '"/a"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/3')

        self.assertEqual(doc, {
            'a': [1, 2],
        })

    def test_replace_top_level_dict_with_array(self):
        """Testing json_patch with replace op replacing top-level dictionary
        with an array
        """
        doc = {}

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'replace',
                    'path': '',
                    'value': [],
                }]),
            [])
        self.assertEqual(doc, {})

    def test_replace_top_level_array_with_dict(self):
        """Testing json_patch with replace op replacing top-level array with
        a dictionary
        """
        doc = []

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'replace',
                    'path': '',
                    'value': {},
                }]),
            {})
        self.assertEqual(doc, [])

    def test_replace_with_dict_key(self):
        """Testing json_patch with replace op replacing key in dictionary"""
        doc = {
            'a': {
                'b': 1,
            },
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'replace',
                    'path': '/a/b',
                    'value': 2,
                }]),
            {
                'a': {
                    'b': 2,
                },
            })
        self.assertEqual(doc, {
            'a': {
                'b': 1,
            },
        })

    def test_replace_with_array_index_0(self):
        """Testing json_patch with replace op with array index 0"""
        doc = {
            'a': {
                'b': [1, 2, 3],
            },
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'replace',
                    'path': '/a/b/0',
                    'value': 100,
                }]),
            {
                'a': {
                    'b': [100, 2, 3],
                },
            })
        self.assertEqual(doc, {
            'a': {
                'b': [1, 2, 3],
            },
        })

    def test_replace_with_array_index_middle(self):
        """Testing json_patch with replace op with array index (middle of
        array)
        """
        doc = {
            'a': {
                'b': [1, 2, 3],
            },
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'replace',
                    'path': '/a/b/1',
                    'value': 100,
                }]),
            {
                'a': {
                    'b': [1, 100, 3],
                },
            })
        self.assertEqual(doc, {
            'a': {
                'b': [1, 2, 3],
            },
        })

    def test_replace_with_array_index_last(self):
        """Testing json_patch with replace op with array index (last in array)
        """
        doc = {
            'a': {
                'b': [1, 2, 3],
            },
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'replace',
                    'path': '/a/b/2',
                    'value': 100,
                }]),
            {
                'a': {
                    'b': [1, 2, 100],
                },
            })
        self.assertEqual(doc, {
            'a': {
                'b': [1, 2, 3],
            },
        })

    def test_replace_with_array_index_end(self):
        """Testing json_patch with replace op with array index (end of array)
        """
        doc = {
            'a': {
                'b': [1, 2, 3],
            },
        }
        patch = [{
            'op': 'replace',
            'path': '/a/b/-',
            'value': 100,
        }]

        message = (
            'Cannot perform operation "replace" on end of list at "/a/b/-" '
            'for patch entry 0'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/b/-')

        self.assertEqual(doc, {
            'a': {
                'b': [1, 2, 3],
            },
        })

    def test_replace_with_can_write_key_func_and_allowed(self):
        """Testing json_patch with replace op and can_write_key_func and key
        allowed
        """
        def _can_write_key(doc, patch_entry, path):
            return path == ('a', 'b')

        doc = {
            'a': {
                'b': 1,
            },
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'replace',
                    'path': '/a/b',
                    'value': 100,
                }],
                can_write_key_func=_can_write_key),
            {
                'a': {
                    'b': 100,
                },
            })

    def test_replace_with_can_write_key_func_and_not_allowed(self):
        """Testing json_patch with replace op and can_write_key_func and key
        not allowed
        """
        def _can_write_key(doc, patch_entry, path):
            return path == ('a', 'c')

        doc = {
            'a': {
                'b': 1,
            },
        }
        patch = [{
            'op': 'replace',
            'path': '/a/b',
            'value': 100,
        }]

        message = 'Cannot write to path "/a/b" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchWriteAccessError,
                                      message) as cm:
            json_patch(doc=doc,
                       patch=patch,
                       can_write_key_func=_can_write_key)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/b')

    def test_replace_with_missing_path(self):
        """Testing json_patch with replace op and missing path"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'replace',
            'value': 1,
        }]

        message = 'Missing key "path" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_replace_with_bad_path(self):
        """Testing json_patch with replace op and bad path"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'replace',
            'path': '/b',
            'value': 100,
        }]

        message = \
            'Cannot remove non-existent key "b" in path "/b" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/b')

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_replace_with_invalid_path_syntax(self):
        """Testing json_patch with replace op and invalid path syntax"""
        doc = {}
        patch = [{
            'op': 'replace',
            'path': 'b',
            'value': 100,
        }]

        message = (
            'Syntax error in path "b" for patch entry 0: Paths must either '
            'be empty or start with a "/"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, 'b')

    def test_copy_with_dict_key(self):
        """Testing json_patch with copy op and dictionary key"""
        doc = {
            'a': 1,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'copy',
                    'from': '/a',
                    'path': '/b'
                }]),
            {
                'a': 1,
                'b': 1,
            })
        self.assertEqual(doc, {
            'a': 1,
        })

    def test_copy_with_array_index(self):
        """Testing json_patch with copy op and array index"""
        doc = {
            'a': [1, 2, 3],
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'copy',
                    'from': '/a/1',
                    'path': '/a/0'
                }]),
            {
                'a': [2, 1, 2, 3],
            })
        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_copy_with_array_index_end(self):
        """Testing json_patch with copy op and array index `-` (end of array)
        """
        doc = {
            'a': [1, 2, 3],
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'copy',
                    'from': '/a/1',
                    'path': '/a/-'
                }]),
            {
                'a': [1, 2, 3, 2],
            })
        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_copy_with_from_array_index_end(self):
        """Testing json_patch with copy op and "from" array index `-` (end of
        array)
        """
        doc = {
            'a': [1, 2, 3],
        }
        patch = [{
            'op': 'copy',
            'from': '/a/-',
            'path': '/a/0'
        }]

        message = (
            'Cannot perform operation "copy" from end of list at "/a/-" for '
            'patch entry 0'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/-')

        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_copy_with_array_index_to_dict_key(self):
        """Testing json_patch with copy op and array index copied to dictionary
        key
        """
        doc = {
            'a': [1, 2, 3],
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'copy',
                    'from': '/a/1',
                    'path': '/b'
                }]),
            {
                'a': [1, 2, 3],
                'b': 2,
            })
        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_copy_with_dict_key_to_array_index(self):
        """Testing json_patch with copy op and dictionary key copied to array
        index
        """
        doc = {
            'a': [1, 2, 3],
            'b': 100,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'copy',
                    'from': '/b',
                    'path': '/a/1'
                }]),
            {
                'a': [1, 100, 2, 3],
                'b': 100,
            })
        self.assertEqual(doc, {
            'a': [1, 2, 3],
            'b': 100,
        })

    def test_copy_with_same_paths(self):
        """Testing json_patch with copy op and same source and destination
        paths
        """
        doc = {
            'a': 1,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'copy',
                    'from': '/a',
                    'path': '/a'
                }]),
            {
                'a': 1,
            })
        self.assertEqual(doc, {
            'a': 1,
        })

    def test_copy_with_can_write_key_func_and_allowed(self):
        """Testing json_patch with copy op and can_write_key_func and key
        allowed
        """
        def _can_write_key(doc, patch_entry, path):
            return path == ('b',)

        doc = {
            'a': {
                'b': 1,
            },
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'copy',
                    'from': '/a/b',
                    'path': '/b',
                }],
                can_write_key_func=_can_write_key),
            {
                'a': {
                    'b': 1,
                },
                'b': 1,
            })

    def test_copy_with_can_write_key_func_and_not_allowed(self):
        """Testing json_patch with copy op and can_write_key_func and key
        not allowed
        """
        def _can_write_key(doc, patch_entry, path):
            return path in ('a', 'c')

        doc = {
            'a': {
                'b': 1,
            },
        }
        patch = [{
            'op': 'copy',
            'from': '/a/b',
            'path': '/b',
        }]

        message = 'Cannot write to path "/b" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchWriteAccessError,
                                      message) as cm:
            json_patch(doc=doc,
                       patch=patch,
                       can_write_key_func=_can_write_key)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/b')

    def test_copy_with_can_read_key_func_and_allowed(self):
        """Testing json_patch with copy op and can_read_key_func and key
        allowed
        """
        def _can_read_key(doc, patch_entry, path):
            return path in (('a', 'b'), ('b',))

        doc = {
            'a': {
                'b': 1,
            },
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'copy',
                    'from': '/a/b',
                    'path': '/b',
                }],
                can_read_key_func=_can_read_key),
            {
                'a': {
                    'b': 1,
                },
                'b': 1,
            })

    def test_copy_with_can_read_key_func_and_not_allowed(self):
        """Testing json_patch with copy op and can_read_key_func and key
        not allowed
        """
        def _can_read_key(doc, patch_entry, path):
            return path == ('b',)

        doc = {
            'a': {
                'b': 1,
            },
        }
        patch = [{
            'op': 'copy',
            'from': '/a/b',
            'path': '/b',
        }]

        message = 'Cannot read from path "/a/b" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchReadAccessError, message) as cm:
            json_patch(doc=doc,
                       patch=patch,
                       can_read_key_func=_can_read_key)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/b')

    def test_copy_with_missing_path(self):
        """Testing json_patch with copy op and missing destination path"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'copy',
            'from': '/a',
        }]

        message = 'Missing key "path" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_copy_with_missing_from(self):
        """Testing json_patch with copy op and missing from path"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'copy',
            'path': '/c',
        }]

        message = 'Missing key "from" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_copy_with_bad_from_path(self):
        """Testing json_patch with copy op and bad from path"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'copy',
            'from': '/b',
            'path': '/c',
        }]

        message = (
            'Invalid from path "/b" for patch entry 0: Dictionary key "b" '
            'not found in "/"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/b')

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_copy_with_bad_path(self):
        """Testing json_patch with copy op and bad destination path"""
        doc = {
            'a': 1,
            'b': 2,
        }

        patch = [{
            'op': 'copy',
            'from': '/a',
            'path': '/b/bad',
        }]

        message = (
            'Unable to add key "bad" to a non-dictionary/list in path '
            '"/b/bad" for patch entry 0'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/b/bad')

        self.assertEqual(doc, {
            'a': 1,
            'b': 2,
        })

    def test_copy_with_invalid_from_path_syntax(self):
        """Testing json_patch with copy op and invalid from path syntax"""
        doc = {}
        patch = [{
            'op': 'copy',
            'from': 'b',
            'path': '/b',
        }]

        message = (
            'Syntax error in from path "b" for patch entry 0: Paths must '
            'either be empty or start with a "/"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, 'b')

    def test_copy_with_invalid_path_syntax(self):
        """Testing json_patch with copy op and invalid path syntax"""
        doc = {}
        patch = [{
            'op': 'copy',
            'from': '/b',
            'path': 'b',
        }]

        message = (
            'Syntax error in path "b" for patch entry 0: Paths must '
            'either be empty or start with a "/"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, 'b')

    def test_move_with_dict_key(self):
        """Testing json_patch with move op and dictionary key"""
        doc = {
            'a': 1,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'move',
                    'from': '/a',
                    'path': '/b'
                }]),
            {
                'b': 1,
            })
        self.assertEqual(doc, {
            'a': 1,
        })

    def test_move_with_array_index(self):
        """Testing json_patch with move op and array index"""
        doc = {
            'a': [1, 2, 3],
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'move',
                    'from': '/a/1',
                    'path': '/a/0'
                }]),
            {
                'a': [2, 1, 3],
            })
        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_move_with_array_index_end(self):
        """Testing json_patch with move op and array index `-` (end of array)
        """
        doc = {
            'a': [1, 2, 3],
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'move',
                    'from': '/a/1',
                    'path': '/a/-'
                }]),
            {
                'a': [1, 3, 2],
            })
        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_move_with_from_array_index_end(self):
        """Testing json_patch with move op and "from" array index `-` (end of
        array)
        """
        doc = {
            'a': [1, 2, 3],
        }
        patch = [{
            'op': 'move',
            'from': '/a/-',
            'path': '/a/0'
        }]

        message = (
            'Cannot perform operation "move" from end of list at "/a/-" for '
            'patch entry 0'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/-')

        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_move_with_array_index_to_dict_key(self):
        """Testing json_patch with move op and array index copied to dictionary
        key
        """
        doc = {
            'a': [1, 2, 3],
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'move',
                    'from': '/a/1',
                    'path': '/b'
                }]),
            {
                'a': [1, 3],
                'b': 2,
            })
        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_move_with_dict_key_to_array_index(self):
        """Testing json_patch with move op and dictionary key moved to array
        index
        """
        doc = {
            'a': [1, 2, 3],
            'b': 100,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'move',
                    'from': '/b',
                    'path': '/a/1'
                }]),
            {
                'a': [1, 100, 2, 3],
            })
        self.assertEqual(doc, {
            'a': [1, 2, 3],
            'b': 100,
        })

    def test_move_with_same_paths(self):
        """Testing json_patch with move op and same source and destination
        paths
        """
        doc = {
            'a': 1,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'move',
                    'from': '/a',
                    'path': '/a'
                }]),
            {
                'a': 1,
            })
        self.assertEqual(doc, {
            'a': 1,
        })

    def test_move_with_can_write_key_func_and_allowed(self):
        """Testing json_patch with move op and can_write_key_func and key
        allowed
        """
        def _can_write_key(doc, patch_entry, path):
            return path in (('a', 'b'), ('b',))

        doc = {
            'a': {
                'b': 1,
            },
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'move',
                    'from': '/a/b',
                    'path': '/b',
                }],
                can_write_key_func=_can_write_key),
            {
                'a': {},
                'b': 1,
            })

    def test_move_with_can_write_key_func_and_not_allowed(self):
        """Testing json_patch with move op and can_write_key_func and key
        not allowed
        """
        def _can_write_key(doc, patch_entry, path):
            return path in ('a', 'c')

        doc = {
            'a': {
                'b': 1,
            },
        }
        patch = [{
            'op': 'move',
            'from': '/a/b',
            'path': '/b',
        }]

        message = 'Cannot write to path "/b" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchWriteAccessError,
                                      message) as cm:
            json_patch(doc=doc,
                       patch=patch,
                       can_write_key_func=_can_write_key)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/b')

    def test_move_with_missing_path(self):
        """Testing json_patch with move op and missing destination path"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'move',
            'from': '/a',
        }]

        message = 'Missing key "path" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_move_with_missing_from(self):
        """Testing json_patch with move op and missing from path"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'move',
            'path': '/c',
        }]

        message = 'Missing key "from" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_move_with_bad_from_path(self):
        """Testing json_patch with move op and bad from path"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'move',
            'from': '/b',
            'path': '/c',
        }]

        message = (
            'Invalid from path "/b" for patch entry 0: Dictionary key "b" '
            'not found in "/"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/b')

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_move_with_bad_path(self):
        """Testing json_patch with move op and bad destination path"""
        doc = {
            'a': 1,
            'b': 2,
        }
        patch = [{
            'op': 'move',
            'from': '/a',
            'path': '/b/bad',
        }]

        message = (
            'Unable to add key "bad" to a non-dictionary/list in path '
            '"/b/bad" for patch entry 0'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/b/bad')

        self.assertEqual(doc, {
            'a': 1,
            'b': 2,
        })

    def test_move_with_invalid_from_path_syntax(self):
        """Testing json_patch with move op and invalid from path syntax"""
        doc = {}
        patch = [{
            'op': 'move',
            'from': 'b',
            'path': '/b',
        }]

        message = (
            'Syntax error in from path "b" for patch entry 0: Paths must '
            'either be empty or start with a "/"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, 'b')

    def test_move_with_invalid_path_syntax(self):
        """Testing json_patch with move op and invalid path syntax"""
        doc = {}
        patch = [{
            'op': 'move',
            'from': '/b',
            'path': 'b',
        }]

        message = (
            'Syntax error in path "b" for patch entry 0: Paths must '
            'either be empty or start with a "/"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, 'b')

    def test_move_with_moving_into_child(self):
        """Testing json_patch with move op and attempting to move into child of
        path
        """
        doc = {
            'a': {
                'b': {
                    'c': {},
                },
            },
        }
        patch = [{
            'op': 'move',
            'from': '/a/b',
            'path': '/a/b/c',
        }]

        message = 'Cannot move values into their own children at patch entry 0'

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/b')

        self.assertEqual(doc, {
            'a': {
                'b': {
                    'c': {},
                },
            },
        })

    def test_test_with_matching_value(self):
        """Testing json_patch with test op and matching value"""
        doc = {
            'a': 1,
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'test',
                    'path': '/a',
                    'value': 1,
                }]),
            {
                'a': 1,
            })

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_test_with_missing_path(self):
        """Testing json_patch with test op and missing path"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'test',
            'value': 1,
        }]

        message = 'Missing key "path" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_test_with_array_index_end(self):
        """Testing json_patch with test op and array index `-` (end of array)
        """
        doc = {
            'a': [1, 2, 3],
        }
        patch = [{
            'op': 'test',
            'path': '/a/-',
            'value': 1,
        }]

        message = (
            'Cannot perform operation "test" on end of list at "/a/-" for '
            'patch entry 0'
        )

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/-')

        self.assertEqual(doc, {
            'a': [1, 2, 3],
        })

    def test_test_with_missing_value(self):
        """Testing json_patch with test op and missing value"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'test',
            'path': '/a',
        }]

        message = 'Missing key "value" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_test_without_matching_value(self):
        """Testing json_patch with test op and non-matching value"""
        doc = {
            'a': 1,
        }
        patch = [{
            'op': 'test',
            'path': '/a',
            'value': 2,
        }]

        message = \
            'Test failed for path "/a" at patch entry 0. Expected 2 and got 1.'

        with self.assertRaisesMessage(JSONPatchTestError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)

        self.assertEqual(doc, {
            'a': 1,
        })

    def test_test_with_can_read_key_func_and_allowed(self):
        """Testing json_patch with test op and can_read_key_func and key
        allowed
        """
        def _can_read_key(doc, patch_entry, path):
            return path == ('a', 'b')

        doc = {
            'a': {
                'b': 1,
            },
        }

        self.assertEqual(
            json_patch(
                doc=doc,
                patch=[{
                    'op': 'test',
                    'path': '/a/b',
                    'value': 1,
                }],
                can_read_key_func=_can_read_key),
            {
                'a': {
                    'b': 1,
                },
            })

    def test_test_with_can_read_key_func_and_not_allowed(self):
        """Testing json_patch with test op and can_read_key_func and key
        not allowed
        """
        def _can_read_key(doc, patch_entry, path):
            return path == ('a', 'c')

        doc = {
            'a': {
                'b': 1,
            },
        }
        patch = [{
            'op': 'test',
            'path': '/a/b',
            'value': 1,
        }]

        message = 'Cannot read from path "/a/b" for patch entry 0'

        with self.assertRaisesMessage(JSONPatchReadAccessError, message) as cm:
            json_patch(doc=doc,
                       patch=patch,
                       can_read_key_func=_can_read_key)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a/b')

    def test_test_with_dict_missing_key(self):
        """Testing json_patch with test op and dictionary with missing key"""
        doc = {}
        patch = [{
            'op': 'test',
            'path': '/a',
            'value': 1,
        }]

        message = (
            'Invalid path "/a" for patch entry 0: Dictionary key "a" not '
            'found in "/"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/a')

        self.assertEqual(doc, {})

    def test_test_with_bad_path(self):
        """Testing json_patch with test op and bad destination path"""
        doc = {
            'a': 1,
            'b': 2,
        }
        patch = [{
            'op': 'test',
            'path': '/b/bad',
            'value': 1,
        }]

        message = 'Cannot resolve path within unsupported type "int" at "/b"'

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, '/b/bad')

        self.assertEqual(doc, {
            'a': 1,
            'b': 2,
        })

    def test_test_with_invalid_path_syntax(self):
        """Testing json_patch with test op and invalid path syntax"""
        doc = {}
        patch = [{
            'op': 'test',
            'path': 'b',
            'value': 1,
        }]

        message = (
            'Syntax error in path "b" for patch entry 0: Paths must '
            'either be empty or start with a "/"'
        )

        with self.assertRaisesMessage(JSONPatchPathError, message) as cm:
            json_patch(doc=doc, patch=patch)

        e = cm.exception
        self.assertEqual(e.doc, doc)
        self.assertEqual(e.patch, patch)
        self.assertEqual(e.patch_entry_index, 0)
        self.assertEqual(e.path, 'b')


class JSONGetPointerInfoTests(TestCase):
    """Unit tests for djblets.util.json_utils.json_get_pointer_info."""

    def test_with_relative_path(self):
        """Testing json_get_pointer_info with relative path"""
        message = 'Paths must either be empty or start with a "/"'

        with self.assertRaisesMessage(JSONPointerSyntaxError, message):
            json_get_pointer_info({}, 'a/')

    def test_with_dict(self):
        """Testing json_get_pointer_info with valid dictionary path"""
        obj = {
            '': 'hi',
            'a': {
                'b': {
                    'c': 123,
                },
            },
        }

        self.assertEqual(
            json_get_pointer_info(obj, ''),
            {
                'value': obj,
                'parent': None,
                'resolved_values': [obj],
                'all_tokens': [],
                'resolved_tokens': [],
                'unresolved_tokens': [],
                'lookup_error': None,
            })
        self.assertEqual(
            json_get_pointer_info(obj, '/'),
            {
                'value': 'hi',
                'parent': obj,
                'resolved_values': [obj, 'hi'],
                'all_tokens': [''],
                'resolved_tokens': [''],
                'unresolved_tokens': [],
                'lookup_error': None,
            })
        self.assertEqual(
            json_get_pointer_info(obj, '/a'),
            {
                'value': obj['a'],
                'parent': obj,
                'resolved_values': [obj, obj['a']],
                'all_tokens': ['a'],
                'resolved_tokens': ['a'],
                'unresolved_tokens': [],
                'lookup_error': None,
            })
        self.assertEqual(
            json_get_pointer_info(obj, '/a/b'),
            {
                'value': obj['a']['b'],
                'parent': obj['a'],
                'resolved_values': [obj,
                                    obj['a'],
                                    obj['a']['b']],
                'all_tokens': ['a', 'b'],
                'resolved_tokens': ['a', 'b'],
                'unresolved_tokens': [],
                'lookup_error': None,
            })
        self.assertEqual(
            json_get_pointer_info(obj, '/a/b/c'),
            {
                'value': obj['a']['b']['c'],
                'parent': obj['a']['b'],
                'resolved_values': [obj,
                                    obj['a'],
                                    obj['a']['b'],
                                    obj['a']['b']['c']],
                'all_tokens': ['a', 'b', 'c'],
                'resolved_tokens': ['a', 'b', 'c'],
                'unresolved_tokens': [],
                'lookup_error': None,
            })

    def test_with_list_item(self):
        """Testing json_get_pointer_info with valid resulting list item"""
        obj = {
            'a': {
                'b': {
                    'c': [1, 2, 3],
                },
            },
        }

        self.assertEqual(
            json_get_pointer_info(obj, '/a/b/c/0'),
            {
                'value': 1,
                'parent': obj['a']['b']['c'],
                'resolved_values': [obj,
                                    obj['a'],
                                    obj['a']['b'],
                                    obj['a']['b']['c'],
                                    obj['a']['b']['c'][0]],
                'all_tokens': ['a', 'b', 'c', '0'],
                'resolved_tokens': ['a', 'b', 'c', '0'],
                'unresolved_tokens': [],
                'lookup_error': None,
            })
        self.assertEqual(
            json_get_pointer_info(obj, '/a/b/c/1'),
            {
                'value': 2,
                'parent': obj['a']['b']['c'],
                'resolved_values': [obj,
                                    obj['a'],
                                    obj['a']['b'],
                                    obj['a']['b']['c'],
                                    obj['a']['b']['c'][1]],
                'all_tokens': ['a', 'b', 'c', '1'],
                'resolved_tokens': ['a', 'b', 'c', '1'],
                'unresolved_tokens': [],
                'lookup_error': None,
            })
        self.assertEqual(
            json_get_pointer_info(obj, '/a/b/c/2'),
            {
                'value': 3,
                'parent': obj['a']['b']['c'],
                'resolved_values': [obj,
                                    obj['a'],
                                    obj['a']['b'],
                                    obj['a']['b']['c'],
                                    obj['a']['b']['c'][2]],
                'all_tokens': ['a', 'b', 'c', '2'],
                'resolved_tokens': ['a', 'b', 'c', '2'],
                'unresolved_tokens': [],
                'lookup_error': None,
            })

    def test_with_list_item_end_of_list(self):
        """Testing json_get_pointer_info with valid resulting list item using
        "-" for end of list
        """
        obj = {
            'a': {
                'b': {
                    'c': [1, 2, 3],
                },
            },
        }

        self.assertEqual(
            json_get_pointer_info(obj, '/a/b/c/-'),
            {
                'value': JSONPointerEndOfList([1, 2, 3]),
                'parent': obj['a']['b']['c'],
                'resolved_values': [obj,
                                    obj['a'],
                                    obj['a']['b'],
                                    obj['a']['b']['c'],
                                    JSONPointerEndOfList([1, 2, 3])],
                'all_tokens': ['a', 'b', 'c', '-'],
                'resolved_tokens': ['a', 'b', 'c', '-'],
                'unresolved_tokens': [],
                'lookup_error': None,
            })

    def test_with_nested_lists_dicts(self):
        """Testing json_get_pointer_info with valid path containing nested
        lists and dictionaries
        """
        obj = {
            'a': [
                1,
                {
                    'b': [2, 3, 'foo'],
                },
            ],
        }

        self.assertEqual(
            json_get_pointer_info(obj, '/a/1/b/2'),
            {
                'value': 'foo',
                'parent': obj['a'][1]['b'],
                'resolved_values': [obj,
                                    obj['a'],
                                    obj['a'][1],
                                    obj['a'][1]['b'],
                                    obj['a'][1]['b'][2]],
                'all_tokens': ['a', '1', 'b', '2'],
                'resolved_tokens': ['a', '1', 'b', '2'],
                'unresolved_tokens': [],
                'lookup_error': None,
            })

    def test_with_escaped(self):
        """Testing json_get_pointer_info with valid escaped paths"""
        obj = {
            '/': {
                '~': {
                    '~/~': 123,
                },
            },
        }

        self.assertEqual(
            json_get_pointer_info(obj, '/~1/~0/~0~1~0'),
            {
                'value': 123,
                'parent': obj['/']['~'],
                'resolved_values': [obj,
                                    obj['/'],
                                    obj['/']['~'],
                                    obj['/']['~']['~/~']],
                'all_tokens': ['/', '~', '~/~'],
                'resolved_tokens': ['/', '~', '~/~'],
                'unresolved_tokens': [],
                'lookup_error': None,
            })

    def test_with_bad_key(self):
        """Testing json_get_pointer_info with invalid path containing a bad
        key
        """
        obj = {
            'a': {
                'b': 123,
            },
        }

        self.assertEqual(
            json_get_pointer_info(obj, '/a/c'),
            {
                'value': None,
                'parent': obj['a'],
                'resolved_values': [obj,
                                    obj['a']],
                'all_tokens': ['a', 'c'],
                'resolved_tokens': ['a'],
                'unresolved_tokens': ['c'],
                'lookup_error': 'Dictionary key "c" not found in "/a"',
            })

    def test_with_bad_list_index(self):
        """Testing json_get_pointer_info with invalid path containing a bad
        list index
        """
        obj = {
            'a': {
                'b': [1, 2],
            },
        }

        self.assertEqual(
            json_get_pointer_info(obj, '/a/b/3'),
            {
                'value': None,
                'parent': obj['a']['b'],
                'resolved_values': [obj,
                                    obj['a'],
                                    obj['a']['b']],
                'all_tokens': ['a', 'b', '3'],
                'resolved_tokens': ['a', 'b'],
                'unresolved_tokens': ['3'],
                'lookup_error': '3 is outside the list in "/a/b"',
            })

    def test_with_non_traversable_path(self):
        """Testing json_get_pointer_info with invalid path containing a
        non-traversable item
        """
        obj = {
            'a': {
                'b': 'test',
            },
        }

        self.assertEqual(
            json_get_pointer_info(obj, '/a/b/c'),
            {
                'value': None,
                'parent': obj['a']['b'],
                'resolved_values': [obj,
                                    obj['a'],
                                    obj['a']['b']],
                'all_tokens': ['a', 'b', 'c'],
                'resolved_tokens': ['a', 'b'],
                'unresolved_tokens': ['c'],
                'lookup_error': (
                    'Cannot resolve path within unsupported type '
                    '"%s" at "/a/b"'
                    % six.text_type.__name__
                ),
            })

    def test_with_negative_list_index(self):
        """Testing json_get_pointer_info with invalid path containing a negative
        list index
        """
        obj = {
            'a': {
                'b': [1, 2],
            },
        }

        message = 'Negative indexes into lists are not allowed'

        with self.assertRaisesMessage(JSONPointerSyntaxError, message):
            json_get_pointer_info(obj, '/a/b/-1')

    def test_with_index_with_leading_zero(self):
        """Testing json_get_pointer_info with invalid path containing a list
        index with a leading zero
        """
        obj = {
            'a': {
                'b': [1, 2],
            },
        }

        message = 'List index "01" must not begin with "0"'

        with self.assertRaisesMessage(JSONPointerSyntaxError, message):
            json_get_pointer_info(obj, '/a/b/01')

    def test_with_non_int_list_index(self):
        """Testing json_get_pointer_info with invalid path containing a
        non-integer list index
        """
        obj = {
            'a': {
                'b': [1, 2],
            },
        }

        message = '%r is not a valid list index in "/a/b"' % 'c'

        with self.assertRaisesMessage(JSONPointerSyntaxError, message):
            json_get_pointer_info(obj, '/a/b/c')


class JSONResolvePointerTests(TestCase):
    """Unit tests for djblets.util.json_utils.json_resolve_pointer."""

    def test_with_dict(self):
        """Testing json_resolve_pointer with valid dictionary path"""
        obj = {
            '': 'hi',
            'a': {
                'b': {
                    'c': 123,
                },
            },
        }

        self.assertEqual(json_resolve_pointer(obj, ''), obj)
        self.assertEqual(json_resolve_pointer(obj, '/'), 'hi')
        self.assertEqual(json_resolve_pointer(obj, '/a'), {
            'b': {
                'c': 123,
            },
        })
        self.assertEqual(json_resolve_pointer(obj, '/a/b'), {
            'c': 123,
        })
        self.assertEqual(json_resolve_pointer(obj, '/a/b/c'), 123)

    def test_with_list_item(self):
        """Testing json_resolve_pointer with valid resulting list item"""
        obj = {
            'a': {
                'b': {
                    'c': [1, 2, 3],
                },
            },
        }

        self.assertEqual(json_resolve_pointer(obj, '/a/b/c/0'), 1)
        self.assertEqual(json_resolve_pointer(obj, '/a/b/c/1'), 2)
        self.assertEqual(json_resolve_pointer(obj, '/a/b/c/2'), 3)

    def test_with_list_item_end_of_list(self):
        """Testing json_resolve_pointer with valid resulting list item using
        "-" for end of list
        """
        obj = {
            'a': {
                'b': {
                    'c': [1, 2, 3],
                },
            },
        }

        self.assertIsInstance(json_resolve_pointer(obj, '/a/b/c/-'),
                              JSONPointerEndOfList)

    def test_with_nested_lists_dicts(self):
        """Testing json_resolve_pointer with valid path containing nested
        lists and dictionaries
        """
        obj = {
            'a': [
                1,
                {
                    'b': [2, 3, 'foo'],
                },
            ],
        }

        self.assertEqual(json_resolve_pointer(obj, '/a/1/b/2'), 'foo')

    def test_with_escaped(self):
        """Testing json_resolve_pointer with valid escaped paths"""
        obj = {
            '/': {
                '~': {
                    '~/~': 123,
                },
            },
        }

        self.assertEqual(json_resolve_pointer(obj, '/~1/~0/~0~1~0'), 123)

    def test_with_bad_key(self):
        """Testing json_resolve_pointer with invalid path containing a bad
        key
        """
        obj = {
            'a': {
                'b': 123,
            },
        }

        message = 'Dictionary key "c" not found in "/a"'

        with self.assertRaisesMessage(JSONPointerLookupError, message):
            json_resolve_pointer(obj, '/a/c')

    def test_with_bad_list_index(self):
        """Testing json_resolve_pointer with invalid path containing a bad
        list index
        """
        obj = {
            'a': {
                'b': [1, 2],
            },
        }

        message = '3 is outside the list in "/a/b"'

        with self.assertRaisesMessage(JSONPointerLookupError, message):
            json_resolve_pointer(obj, '/a/b/3')

    def test_with_non_traversable_path(self):
        """Testing json_resolve_pointer with invalid path containing a
        non-traversable item
        """
        obj = {
            'a': {
                'b': 'test',
            },
        }

        message = (
            'Cannot resolve path within unsupported type "%s" at "/a/b"'
            % six.text_type.__name__
        )

        with self.assertRaisesMessage(JSONPointerLookupError, message):
            json_resolve_pointer(obj, '/a/b/c')

    def test_with_index_with_leading_zero(self):
        """Testing json_resolve_pointer with invalid path containing a negative
        list index
        """
        obj = {
            'a': {
                'b': [1, 2],
            },
        }

        message = 'Negative indexes into lists are not allowed'

        with self.assertRaisesMessage(JSONPointerSyntaxError, message):
            json_resolve_pointer(obj, '/a/b/-1')

    def test_with_negative_list_index(self):
        """Testing json_resolve_pointer with invalid path containing a list
        index with a leading zero
        """
        obj = {
            'a': {
                'b': [1, 2],
            },
        }

        message = 'List index "01" must not begin with "0"'

        with self.assertRaisesMessage(JSONPointerSyntaxError, message):
            json_resolve_pointer(obj, '/a/b/01')

    def test_with_non_int_list_index(self):
        """Testing json_resolve_pointer with invalid path containing a
        non-integer list index
        """
        obj = {
            'a': {
                'b': [1, 2],
            },
        }

        message = '%r is not a valid list index in "/a/b"' % 'c'

        with self.assertRaisesMessage(JSONPointerSyntaxError, message):
            json_resolve_pointer(obj, '/a/b/c')



"""Evolution to remove unique constraint from StoredConsentData."""

from __future__ import unicode_literals

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('StoredConsentData', 'audit_identifier', initial=None,
                unique=False),
]



"""Database backend for MySQL with backported fixes."""

from __future__ import unicode_literals

from django.core.exceptions import ImproperlyConfigured

try:
    from django.db.backends.mysql.base import (DatabaseWrapper as
                                               BaseMySQLDatabaseWrapper)
except ImproperlyConfigured as e:
    # We can't load this backend, but flat-out aborting the module loading now
    # will cause errors in unit test discovery. So we're going to effectively
    # stub this out. It will be useless if actually instantiated.
    BaseMySQLDatabaseWrapper = object


class DatabaseWrapper(BaseMySQLDatabaseWrapper):
    """Database backend for MySQL.

    This is a specialized version of the standard Django MySQL database backend
    which adds backported compatibility fixes from newer versions of Django.

    Currently, this fixes an issue where contents going into a
    :py:class:`~django.db.models.BinaryField` could trigger a MySQL warning
    due to the binary contents being validated as Unicode. This bug was fixed
    in Django 1.10.5, but is present on older versions.

    To use this backend, just use ``djblets.db.backends.mysql`` as the database
    backend instead of ``django.db.backends.mysql`` in :file:`settings.py`.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the database backend.

        Args:
            *args (tuple):
                Positional arguments for the backend.

            **kwargs (dict):
                Keyword arguments for the backend.
        """
        super(DatabaseWrapper, self).__init__(*args, **kwargs)

        # Django 1.10.5 fixes an issue where BinaryField on certain versions
        # of MySQL results in warnings when injecting non-Unicode text. If
        # the database backend doesn't have this fix, we need to inject it.
        ops_cls = self.ops.__class__

        if not hasattr(ops_cls, 'binary_placeholder_sql'):
            from django.db.models.fields import BinaryField

            assert not hasattr(BinaryField, 'get_placeholder')

            ops_cls.binary_placeholder_sql = self._ops_binary_placeholder_sql
            BinaryField.get_placeholder = self._binary_field_get_placeholder

    def _ops_binary_placeholder_sql(self, value):
        """Return the placeholder format string for binary content.

        This is used by :py:class:`~django.db.models.BinaryField` to retrieve
        the format string for injecting binary contents into the database.
        On MySQL, we want to prefix the contents with ``_binary``.

        Args:
            value (bytes):
                The binary contents used to determine what kind of format
                string to use.

        Returns:
            unicode:
            The format string. This is intended to be Unicode, according to
            the Django 1.10.5 code.
        """
        if value is not None:
            return '_binary %s'
        else:
            return '%s'

    def _binary_field_get_placeholder(self, value, *args):
        """Return the placeholder format string used for the BinaryField.

        If the database backend supports it, this will return a special
        format string used for injecting binary content into the database.

        Args:
            value (bytes):
                The binary contents used to determine what kind of format
                string to use.

            *args (tuple):
                Additional arguments passed to this function. Depending on
                the version of Django, this will either encompass one or two
                arguments (the connection being the last one).

        Returns:
            unicode:
            The format string. This is intended to be Unicode, according to
            the Django 1.10.5 code.
        """
        connection = args[-1]
        ops = connection.ops

        if hasattr(ops, 'binary_placeholder_sql'):
            return ops.binary_placeholder_sql(value)
        else:
            return '%s'



"""Legacy template tags for datagrids.

Deprecated::
    2.0:
    This module is now empty and should no longer be used. It is scheduled
    for removal in Djblets 3.0.
"""

from __future__ import unicode_literals

import warnings

from django import template

from djblets.deprecation import RemovedInDjblets30Warning


register = template.Library()


warnings.warn('djblets.datagrid.templatetags is deprecated and will be '
              'removed in Djblets 3.0.',
              RemovedInDjblets30Warning)



from __future__ import unicode_literals

from django.template import Context, Template, TemplateSyntaxError
from django.test import RequestFactory
from kgb import SpyAgency

from djblets.features import Feature, FeatureLevel, get_features_registry
from djblets.testing.testcases import TestCase


class IfFeatureTests(SpyAgency, TestCase):
    """Base class for ``{% if_feature_* %}`` template tag tests."""

    def setUp(self):
        super(IfFeatureTests, self).setUp()

        get_features_registry().reset()

    def tearDown(self):
        super(IfFeatureTests, self).tearDown()

        get_features_registry().reset()


class IfFeatureEnabledTagTests(IfFeatureTests):
    """Unit tests for the ``{% if_feature_enabled %}`` template tag."""

    def test_missing_arguments(self):
        """Testing if_feature_enabled template tag with missing arguments"""
        with self.assertRaises(TemplateSyntaxError):
            Template('{% load features %}'
                     '{% if_feature_enabled %}'
                     'enabled'
                     '{% endif_feature_enabled %}')

    def test_extra_positional_arguments(self):
        """Testing if_feature_enabled template tag with extra positional
        arguments
        """
        with self.assertRaises(TemplateSyntaxError):
            Template('{% load features %}'
                     '{% if_feature_enabled "my-feature" bar %}'
                     'enabled'
                     '{% endif_feature_enabled %}')

    def test_keyword_arguments(self):
        """Testing if_feature_enabled template tag with keyword arguments"""
        class DummyFeature(Feature):
            feature_id = 'my-feature'
            level = FeatureLevel.STABLE

        feature = DummyFeature()
        self.spy_on(feature.is_enabled)

        t = Template(
            '{% load features %}'
            '{% if_feature_enabled "my-feature" i=42 s="abc" b=True %}'
            'enabled'
            '{% endif_feature_enabled %}')

        request = RequestFactory().get('/')

        self.assertEqual(
            t.render(Context({
                'request': request,
            })),
            'enabled')
        self.assertTrue(feature.is_enabled.called_with(
            request=request,
            i=42,
            s='abc',
            b=True))

    def test_enabled(self):
        """Testing if_feature_enabled template tag with an enabled feature"""
        class DummyFeature(Feature):
            feature_id = 'my-feature'
            level = FeatureLevel.STABLE

        DummyFeature()

        t = Template('{% load features %}'
                     '{% if_feature_enabled "my-feature" %}'
                     'enabled'
                     '{% endif_feature_enabled %}')

        self.assertEqual(t.render(Context()), 'enabled')

    def test_disabled(self):
        """Testing if_feature_enabled template tag with a disabled feature"""
        class DummyFeature(Feature):
            feature_id = 'my-feature'
            level = FeatureLevel.BETA

        DummyFeature()

        t = Template('{% load features %}'
                     '{% if_feature_enabled "my-feature" %}'
                     'enabled'
                     '{% endif_feature_enabled %}')

        self.assertEqual(t.render(Context()), '')

    def test_invalid_feature_id(self):
        """Testing if_feature_enabled template tag with an invalid feature ID
        """
        t = Template('{% load features %}'
                     '{% if_feature_enabled "my-feature" %}'
                     'enabled'
                     '{% endif_feature_enabled %}')

        self.assertEqual(t.render(Context()), '')

    def test_else_with_enabled(self):
        """Testing if_feature_enabled template tag with else block and an
        enabled feature
        """
        class DummyFeature(Feature):
            feature_id = 'my-feature'
            level = FeatureLevel.STABLE

        DummyFeature()

        t = Template('{% load features %}'
                     '{% if_feature_enabled "my-feature" %}'
                     'enabled'
                     '{% else %}'
                     'disabled'
                     '{% endif_feature_enabled %}')

        self.assertEqual(t.render(Context()), 'enabled')

    def test_else_with_disabled(self):
        """Testing if_feature_enabled template tag with else block and a
        disabled feature
        """
        class DummyFeature(Feature):
            feature_id = 'my-feature'
            level = FeatureLevel.BETA

        DummyFeature()

        t = Template('{% load features %}'
                     '{% if_feature_enabled "my-feature" %}'
                     'enabled'
                     '{% else %}'
                     'disabled'
                     '{% endif_feature_enabled %}')

        self.assertEqual(t.render(Context()), 'disabled')

    def test_else_with_invalid_feature_id(self):
        """Testing if_feature_enabled template tag with else block and an
        invalid feature ID
        """
        t = Template('{% load features %}'
                     '{% if_feature_enabled "my-feature" %}'
                     'enabled'
                     '{% else %}'
                     'disabled'
                     '{% endif_feature_enabled %}')

        self.assertEqual(t.render(Context()), 'disabled')


class IfFeatureDisabledTagTests(IfFeatureTests):
    """Unit tests for the ``{% if_feature_disabled %}`` template tag."""

    def test_missing_arguments(self):
        """Testing if_feature_disabled template tag with missing arguments"""
        with self.assertRaises(TemplateSyntaxError):
            Template('{% load features %}'
                     '{% if_feature_disabled %}'
                     'enabled'
                     '{% endif_feature_disabled %}')

    def test_extra_positional_arguments(self):
        """Testing if_feature_disabled template tag with extra positional
        arguments
        """
        with self.assertRaises(TemplateSyntaxError):
            Template('{% load features %}'
                     '{% if_feature_disabled "my-feature" bar %}'
                     'enabled'
                     '{% endif_feature_disabled %}')

    def test_keyword_arguments(self):
        """Testing if_feature_disabled template tag with keyword arguments"""
        class DummyFeature(Feature):
            feature_id = 'my-feature'
            level = FeatureLevel.BETA

        feature = DummyFeature()
        self.spy_on(feature.is_enabled)

        t = Template(
            '{% load features %}'
            '{% if_feature_disabled "my-feature" i=42 s="abc" b=True %}'
            'disabled'
            '{% endif_feature_disabled %}')

        request = RequestFactory().get('/')

        self.assertEqual(
            t.render(Context({
                'request': request,
            })),
            'disabled')
        self.assertTrue(feature.is_enabled.called_with(
            request=request,
            i=42,
            s='abc',
            b=True))

    def test_enabled(self):
        """Testing if_feature_disabled template tag with an enabled feature"""
        class DummyFeature(Feature):
            feature_id = 'my-feature'
            level = FeatureLevel.STABLE

        DummyFeature()

        t = Template('{% load features %}'
                     '{% if_feature_disabled "my-feature" %}'
                     'disabled'
                     '{% endif_feature_disabled %}')

        self.assertEqual(t.render(Context()), '')

    def test_disabled(self):
        """Testing if_feature_disabled template tag with a disabled feature"""
        class DummyFeature(Feature):
            feature_id = 'my-feature'
            level = FeatureLevel.BETA

        DummyFeature()

        t = Template('{% load features %}'
                     '{% if_feature_disabled "my-feature" %}'
                     'disabled'
                     '{% endif_feature_disabled %}')

        self.assertEqual(t.render(Context()), 'disabled')

    def test_invalid_feature_id(self):
        """Testing if_feature_disabled template tag with an invalid feature ID
        """
        t = Template('{% load features %}'
                     '{% if_feature_disabled "my-feature" %}'
                     'disabled'
                     '{% endif_feature_disabled %}')

        self.assertEqual(t.render(Context()), 'disabled')

    def test_else_with_enabled(self):
        """Testing if_feature_disabled template tag with else block and an
        enabled feature
        """
        class DummyFeature(Feature):
            feature_id = 'my-feature'
            level = FeatureLevel.STABLE

        DummyFeature()

        t = Template('{% load features %}'
                     '{% if_feature_disabled "my-feature" %}'
                     'disabled'
                     '{% else %}'
                     'enabled'
                     '{% endif_feature_disabled %}')

        self.assertEqual(t.render(Context()), 'enabled')

    def test_else_with_disabled(self):
        """Testing if_feature_disabled template tag with else block and a
        disabled feature
        """
        class DummyFeature(Feature):
            feature_id = 'my-feature'
            level = FeatureLevel.BETA

        DummyFeature()

        t = Template('{% load features %}'
                     '{% if_feature_disabled "my-feature" %}'
                     'disabled'
                     '{% else %}'
                     'enabled'
                     '{% endif_feature_disabled %}')

        self.assertEqual(t.render(Context()), 'disabled')

    def test_else_with_invalid_feature_id(self):
        """Testing if_feature_disabled template tag with else block and an
        invalid feature ID
        """
        t = Template('{% load features %}'
                     '{% if_feature_disabled "my-feature" %}'
                     'disabled'
                     '{% else %}'
                     'enabled'
                     '{% endif_feature_disabled %}')

        self.assertEqual(t.render(Context()), 'disabled')



PIPELINE_JAVASCRIPT = {
    'djblets-avatars-config': {
        'source_filenames': (
            'djblets/js/avatars/base.js',
            'djblets/js/avatars/models/avatarSettingsModel.es6.js',
            'djblets/js/avatars/views/avatarServiceSettingsFormView.es6.js',
            'djblets/js/avatars/views/avatarSettingsFormView.es6.js',
            'djblets/js/avatars/views/fileUploadSettingsFormView.es6.js',
        ),
        'output_filename': 'djblets/js/avatars-config.min.js',
    },
    'djblets-config-forms': {
        'source_filenames': (
            'djblets/js/configForms/base.js',
            'djblets/js/configForms/collections/listItemsCollection.es6.js',
            'djblets/js/configForms/models/listItemModel.es6.js',
            'djblets/js/configForms/models/listModel.js',
            'djblets/js/configForms/views/listItemView.es6.js',
            'djblets/js/configForms/views/listView.es6.js',
            'djblets/js/configForms/views/pagesView.es6.js',
            'djblets/js/configForms/views/tableItemView.es6.js',
            'djblets/js/configForms/views/tableView.es6.js',
        ),
        'output_filename': 'djblets/js/config-forms.min.js',
    },
    'djblets-datagrid': {
        'source_filenames': ('djblets/js/datagrid.js',),
        'output_filename': 'djblets/js/datagrid.min.js',
    },
    'djblets-extensions-admin': {
        'source_filenames': (
            'djblets/js/extensions/models/extensionManagerModel.es6.js',
            'djblets/js/extensions/views/extensionManagerView.es6.js',
        ),
        'output_filename': 'djblets/js/extensions-admin.min.js',
    },
    'djblets-extensions': {
        'source_filenames': (
            'djblets/js/extensions/base.js',
            'djblets/js/extensions/models/extensionModel.es6.js',
            'djblets/js/extensions/models/extensionHookModel.es6.js',
            'djblets/js/extensions/models/extensionHookPointModel.es6.js',
        ),
        'output_filename': 'djblets/js/extensions.min.js',
    },
    'djblets-forms': {
        'source_filenames': (
            'djblets/js/forms/base.js',
            'djblets/js/forms/models/conditionChoiceModel.es6.js',
            'djblets/js/forms/models/conditionModel.es6.js',
            'djblets/js/forms/models/conditionOperatorModel.es6.js',
            'djblets/js/forms/models/conditionSetModel.es6.js',
            'djblets/js/forms/models/conditionValueField.es6.js',
            'djblets/js/forms/views/baseConditionValueFieldView.es6.js',
            'djblets/js/forms/views/conditionSetView.es6.js',
            'djblets/js/forms/views/conditionValueFormFieldView.es6.js',
            'djblets/js/forms/views/listEditView.es6.js',
            'djblets/js/forms/views/privacyConsentFieldView.es6.js',
        ),
        'output_filename': 'djblets/js/forms.min.js',
    },
    'djblets-integrations': {
        'source_filenames': (
            'djblets/js/integrations/base.js',
            'djblets/js/integrations/views/addIntegrationPopupView.es6.js',
            'djblets/js/integrations/views/integrationConfigListView.es6.js',
        ),
        'output_filename': 'djblets/js/integrations.min.js',
    },
    'djblets-gravy': {
        'source_filenames': (
            # These are in dependency order
            'djblets/js/jquery.gravy.util.js',
            'djblets/js/jquery.gravy.retina.js',
            'djblets/js/jquery.gravy.autosize.js',
            'djblets/js/jquery.gravy.inlineEditor.js',
            'djblets/js/jquery.gravy.modalBox.js',
            'djblets/js/jquery.gravy.tooltip.js',
            'djblets/js/jquery.gravy.funcQueue.js',
            'djblets/js/jquery.gravy.backboneUtils.js',
        ),
        'output_filename': 'djblets/js/jquery.gravy.min.js',
    },
    'djblets-js-tests': {
        'source_filenames': (
            'djblets/js/tests/backboneUtilsTests.js',
            'djblets/js/configForms/models/tests/listItemModelTests.es6.js',
            'djblets/js/configForms/views/tests/listItemViewTests.es6.js',
            'djblets/js/configForms/views/tests/listViewTests.es6.js',
            'djblets/js/configForms/views/tests/tableItemViewTests.es6.js',
            'djblets/js/configForms/views/tests/tableViewTests.es6.js',
            'djblets/js/forms/models/tests/conditionChoiceModelTests.es6.js',
            'djblets/js/forms/models/tests/conditionModelTests.es6.js',
            'djblets/js/forms/models/tests/conditionOperatorModelTests.es6.js',
            'djblets/js/forms/models/tests/conditionSetModelTests.es6.js',
            'djblets/js/forms/views/tests/conditionSetViewTests.es6.js',
            'djblets/js/forms/views/tests/conditionValueFormFieldViewTests.es6.js',
            'djblets/js/forms/views/tests/listEditViewTests.es6.js',
            'djblets/js/integrations/views/tests/addIntegrationPopupViewTests.es6.js',
            'djblets/js/integrations/views/tests/integrationConfigListViewTests.es6.js',
            'djblets/js/utils/tests/urlsTests.es6.js',
        ),
        'output_filename': 'djblets/js/tests.min.js',
    },
    'djblets-utils': {
        'source_filenames': (
            'djblets/js/utils/promise.es6.js',
            'djblets/js/utils/urls.es6.js',
        ),
        'output_filename': 'djblets/js/utils.min.js',
    },
    'djblets-widgets': {
        'source_filenames': (
            'djblets/js/admin/views/relatedObjectSelectorView.es6.js',
            'lib/js/selectize-0.12.4.js',
        ),
        'output_filename': 'djblets/js/widgets.min.js',
    },
}


PIPELINE_STYLESHEETS = {
    'djblets-avatars-config': {
        'source_filenames': (
            'djblets/css/avatars.less',
        ),
        'output_filename': 'djblets/css/avatars-config.min.css',
    },
    'djblets-admin': {
        'source_filenames': (
            'djblets/css/admin.less',
            'djblets/css/extensions.less',
        ),
        'output_filename': 'djblets/css/admin.min.css',
    },
    'djblets-forms': {
        'source_filenames': (
            'djblets/css/forms/conditions.less',
            'djblets/css/forms/copyable_text_input.less',
            'djblets/css/forms/list_edit.less',
            'djblets/css/forms/privacy.less',
        ),
        'output_filename': 'djblets/css/forms.min.css',
    },
    'djblets-config-forms': {
        'source_filenames': (
            'djblets/css/config-forms.less',
        ),
        'output_filename': 'djblets/css/config-forms.min.css',
    },
    'djblets-datagrid': {
        'source_filenames': (
            'djblets/css/datagrid.less',
        ),
        'output_filename': 'djblets/css/datagrid.min.css',
    },
    'djblets-integrations': {
        'source_filenames': (
            'djblets/css/integrations.less',
        ),
        'output_filename': 'djblets/css/integrations.min.css',
    },
    'djblets-ui': {
        'source_filenames': (
            'lib/css/selectize.default-0.12.4.css',
            'djblets/css/ui/modalbox.less',
            'djblets/css/ui/related-object-selector.less',
            'djblets/css/ui/spinner.less',
        ),
        'output_filename': 'djblets/css/ui.min.css',
        'absolute_paths': False,
    },
}



"""Unit tests for ETagViewMixin."""

from __future__ import unicode_literals

from django.http import HttpResponse, HttpResponseNotModified
from django.test.client import RequestFactory
from django.views.generic.base import View

from djblets.testing.testcases import TestCase
from djblets.util.http import encode_etag
from djblets.views.generic.etag import ETagViewMixin


class ETagViewMixinTests(TestCase):
    """Unit tests for ETagViewMixin."""

    def test_get_with_matching_etag(self):
        """Testing ETagViewMixin.get with matching ETag"""
        self._test_with_matching_etag('GET')

    def test_get_without_matching_etag(self):
        """Testing ETagViewMixin.get without matching ETag"""
        self._test_without_matching_etag('GET')

    def test_head_with_matching_etag(self):
        """Testing ETagViewMixin.head with matching ETag"""
        self._test_with_matching_etag('HEAD')

    def test_head_without_matching_etag(self):
        """Testing ETagViewMixin.head without matching ETag"""
        self._test_without_matching_etag('HEAD')

    def test_post_ignores_etags(self):
        """Testing ETagViewMixin.post ignores ETags"""
        self._test_ignore_etag('POST')

    def test_put_ignores_etags(self):
        """Testing ETagViewMixin.put ignores ETags"""
        self._test_ignore_etag('PUT')

    def test_patch_ignores_etags(self):
        """Testing ETagViewMixin.patch ignores ETags"""
        self._test_ignore_etag('PATCH')

    def test_delete_ignores_etags(self):
        """Testing ETagViewMixin.delete ignores ETags"""
        self._test_ignore_etag('DELETE')

    def _test_with_matching_etag(self, method):
        class MyView(ETagViewMixin, View):
            def get_etag_data(self, *args, **kwargs):
                return 'test123'

        setattr(MyView, method.lower(), lambda *args, **kwargs: HttpResponse())

        view = MyView.as_view()
        request = RequestFactory().request(REQUEST_METHOD='HEAD')
        request.META['HTTP_IF_NONE_MATCH'] = encode_etag('test123')

        response = view(request)
        self.assertIsInstance(response, HttpResponseNotModified)
        self.assertFalse(response.has_header('ETag'))

    def _test_without_matching_etag(self, method):
        class MyView(ETagViewMixin, View):
            def get_etag_data(self, *args, **kwargs):
                return 'test123'

            def head(self, *args, **kwargs):
                return HttpResponse()

        setattr(MyView, method.lower(), lambda *args, **kwargs: HttpResponse())

        view = MyView.as_view()
        request = RequestFactory().request(REQUEST_METHOD=method)
        request.META['HTTP_IF_NONE_MATCH'] = encode_etag('nope')

        response = view(request)
        self.assertNotIsInstance(response, HttpResponseNotModified)
        self.assertTrue(response.has_header('ETag'))
        self.assertEqual(response['ETag'], encode_etag('test123'))

    def _test_ignore_etag(self, method):
        class MyView(ETagViewMixin, View):
            def get_etag_data(self, *args, **kwargs):
                return 'test123'

        setattr(MyView, method.lower(),
                lambda *args, **kwargs: HttpResponse('hi there'))

        view = MyView.as_view()
        request = RequestFactory().request(REQUEST_METHOD=method)
        request.META['HTTP_IF_NONE_MATCH'] = encode_etag('test123')

        response = view(request)
        self.assertNotIsInstance(response, HttpResponseNotModified)
        self.assertEqual(response.content, b'hi there')
        self.assertFalse(response.has_header('ETag'))



"""Backports of functionality from Django 1.11's django.utils.functional.

Note that this module is not considered API-stable. Changes may be made that
remove or alter functionality in the future. Please use at your own risk.
"""

# TODO: Remove this file once we no longer support a version of Django
#       prior to 1.11.x.
#
# Much of this is a subset of Django 1.11.16's django/utils/functional.py.
#
# Copyright (c) Django Software Foundation and individual contributors.  All
# rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
#     3. Neither the name of Django nor the names of its contributors may be
#     used to endorse or promote products derived from this software without
#     specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals

import copy
import operator

import django
from django.utils import six
from django.utils.functional import empty, new_method_proxy


if django.VERSION[:3] >= (1, 11, 16):
    from django.utils.functional import (LazyObject, SimpleLazyObject,
                                         unpickle_lazyobject)
else:
    class LazyObject(object):
        """
        A wrapper for another class that can be used to delay instantiation of the
        wrapped class.

        By subclassing, you have the opportunity to intercept and alter the
        instantiation. If you don't need to do that, use SimpleLazyObject.
        """

        # Avoid infinite recursion when tracing __init__ (#19456).
        _wrapped = None

        def __init__(self):
            # Note: if a subclass overrides __init__(), it will likely need to
            # override __copy__() and __deepcopy__() as well.
            self._wrapped = empty

        __getattr__ = new_method_proxy(getattr)

        def __setattr__(self, name, value):
            if name == "_wrapped":
                # Assign to __dict__ to avoid infinite __setattr__ loops.
                self.__dict__["_wrapped"] = value
            else:
                if self._wrapped is empty:
                    self._setup()
                setattr(self._wrapped, name, value)

        def __delattr__(self, name):
            if name == "_wrapped":
                raise TypeError("can't delete _wrapped.")
            if self._wrapped is empty:
                self._setup()
            delattr(self._wrapped, name)

        def _setup(self):
            """
            Must be implemented by subclasses to initialize the wrapped object.
            """
            raise NotImplementedError('subclasses of LazyObject must provide a _setup() method')

        # Because we have messed with __class__ below, we confuse pickle as to what
        # class we are pickling. We're going to have to initialize the wrapped
        # object to successfully pickle it, so we might as well just pickle the
        # wrapped object since they're supposed to act the same way.
        #
        # Unfortunately, if we try to simply act like the wrapped object, the ruse
        # will break down when pickle gets our id(). Thus we end up with pickle
        # thinking, in effect, that we are a distinct object from the wrapped
        # object, but with the same __dict__. This can cause problems (see #25389).
        #
        # So instead, we define our own __reduce__ method and custom unpickler. We
        # pickle the wrapped object as the unpickler's argument, so that pickle
        # will pickle it normally, and then the unpickler simply returns its
        # argument.
        def __reduce__(self):
            if self._wrapped is empty:
                self._setup()
            return (unpickle_lazyobject, (self._wrapped,))

        # Overriding __class__ stops __reduce__ from being called on Python 2.
        # So, define __getstate__ in a way that cooperates with the way that
        # pickle interprets this class. This fails when the wrapped class is a
        # builtin, but it's better than nothing.
        def __getstate__(self):
            if self._wrapped is empty:
                self._setup()
            return self._wrapped.__dict__

        def __copy__(self):
            if self._wrapped is empty:
                # If uninitialized, copy the wrapper. Use type(self), not
                # self.__class__, because the latter is proxied.
                return type(self)()
            else:
                # If initialized, return a copy of the wrapped object.
                return copy.copy(self._wrapped)

        def __deepcopy__(self, memo):
            if self._wrapped is empty:
                # We have to use type(self), not self.__class__, because the
                # latter is proxied.
                result = type(self)()
                memo[id(self)] = result
                return result
            return copy.deepcopy(self._wrapped, memo)

        if six.PY3:
            __bytes__ = new_method_proxy(bytes)
            __str__ = new_method_proxy(str)
            __bool__ = new_method_proxy(bool)
        else:
            __str__ = new_method_proxy(str)
            __unicode__ = new_method_proxy(unicode)  # NOQA: unicode undefined on PY3
            __nonzero__ = new_method_proxy(bool)

        # Introspection support
        __dir__ = new_method_proxy(dir)

        # Need to pretend to be the wrapped class, for the sake of objects that
        # care about this (especially in equality tests)
        __class__ = property(new_method_proxy(operator.attrgetter("__class__")))
        __eq__ = new_method_proxy(operator.eq)
        __ne__ = new_method_proxy(operator.ne)
        __hash__ = new_method_proxy(hash)

        # List/Tuple/Dictionary methods support
        __getitem__ = new_method_proxy(operator.getitem)
        __setitem__ = new_method_proxy(operator.setitem)
        __delitem__ = new_method_proxy(operator.delitem)
        __iter__ = new_method_proxy(iter)
        __len__ = new_method_proxy(len)
        __contains__ = new_method_proxy(operator.contains)


    class SimpleLazyObject(LazyObject):
        """
        A lazy object initialized from any function.

        Designed for compound objects of unknown type. For builtins or objects of
        known type, use django.utils.functional.lazy.
        """
        def __init__(self, func):
            """
            Pass in a callable that returns the object to be wrapped.

            If copies are made of the resulting SimpleLazyObject, which can happen
            in various circumstances within Django, then you must ensure that the
            callable can be safely run more than once and will return the same
            value.
            """
            self.__dict__['_setupfunc'] = func
            super(SimpleLazyObject, self).__init__()

        def _setup(self):
            self._wrapped = self._setupfunc()

        # Return a meaningful representation of the lazy object for debugging
        # without evaluating the wrapped object.
        def __repr__(self):
            if self._wrapped is empty:
                repr_attr = self._setupfunc
            else:
                repr_attr = self._wrapped
            return '<%s: %r>' % (type(self).__name__, repr_attr)

        def __copy__(self):
            if self._wrapped is empty:
                # If uninitialized, copy the wrapper. Use SimpleLazyObject, not
                # self.__class__, because the latter is proxied.
                return SimpleLazyObject(self._setupfunc)
            else:
                # If initialized, return a copy of the wrapped object.
                return copy.copy(self._wrapped)

        def __deepcopy__(self, memo):
            if self._wrapped is empty:
                # We have to use SimpleLazyObject, not self.__class__, because the
                # latter is proxied.
                result = SimpleLazyObject(self._setupfunc)
                memo[id(self)] = result
                return result
            return copy.deepcopy(self._wrapped, memo)


    def unpickle_lazyobject(wrapped):
        """
        Used to unpickle lazy objects. Just return its argument, which will be the
        wrapped object.
        """
        return wrapped


__all__ = (
    'LazyObject',
    'SimpleLazyObject',
    'unpickle_lazyobject',
)



from __future__ import unicode_literals

import os

from django.conf.urls import url

from djblets.feedview.views import view_feed


FEED_URL = "file://%s/testdata/sample.rss" % os.path.dirname(__file__)


urlpatterns = [
    url(r'^feed/$', view_feed, {
        'template_name': 'feedview/feed-page.html',
        'url': FEED_URL
    }),
    url(r'^feed-inline/$', view_feed, {
        'template_name': 'feedview/feed-inline.html',
        'url': FEED_URL
     }),
    url(r'^feed-error/$', view_feed, {
        'template_name':
        'feedview/feed-inline.html',
        'url': 'http://example.fake/dummy.rss'
    }),
]



#
# Settings for djblets.
#
# This is meant for internal use only. We use it primarily for building
# static media to bundle with djblets.
#
# This should generally not be used in a project.
from __future__ import unicode_literals

import os

from djblets.pipeline.settings import build_pipeline_settings
from djblets.staticbundles import PIPELINE_JAVASCRIPT, PIPELINE_STYLESHEETS


SECRET_KEY = '47157c7ae957f904ab809d8c5b77e0209221d4c0'

USE_I18N = True

DEBUG = False
DJBLETS_ROOT = os.path.abspath(os.path.dirname(__file__))
HTDOCS_ROOT = os.path.join(DJBLETS_ROOT, 'htdocs')
STATIC_ROOT = os.path.join(HTDOCS_ROOT, 'static')
STATIC_URL = '/'
LOGIN_LIMIT_RATE = '5/m'

STATICFILES_DIRS = (
    os.path.join(DJBLETS_ROOT, 'static'),
)

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
)

STATICFILES_STORAGE = 'pipeline.storage.PipelineCachedStorage'


NODE_PATH = os.path.abspath(os.path.join(DJBLETS_ROOT, '..', 'node_modules'))


PIPELINE = build_pipeline_settings(
    pipeline_enabled=not DEBUG or os.getenv('FORCE_BUILD_MEDIA'),
    node_modules_path=NODE_PATH,
    static_root=STATIC_ROOT,
    javascript_bundles=PIPELINE_JAVASCRIPT,
    stylesheet_bundles=PIPELINE_STYLESHEETS,
    validate_paths=DEBUG)


INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sites',
    'django.contrib.staticfiles',
    'djblets.auth',
    'djblets.datagrid',
    'djblets.extensions',
    'djblets.feedview',
    'djblets.gravatars',
    'djblets.log',
    'djblets.pipeline',
    'djblets.privacy',
    'djblets.siteconfig',
    'djblets.testing',
    'djblets.util',
    'djblets.webapi',
    'django_evolution',
]



"""Unit tests for djblets.configforms.forms.ConfigPageForm."""

from __future__ import unicode_literals

import warnings

from django import forms
from django.contrib.auth.models import User
from django.test.client import RequestFactory
from django.utils import six

from djblets.configforms.forms import ConfigPageForm
from djblets.configforms.pages import ConfigPage
from djblets.configforms.views import ConfigPagesView
from djblets.testing.testcases import TestCase


class TestForm(ConfigPageForm):
    form_id = 'my-form'
    field1 = forms.CharField(label='Field 1',
                             required=False)
    field2 = forms.CharField(label='Field 2',
                             required=False)


class TestPage(ConfigPage):
    page_id = 'my-page'
    form_classes = [TestForm]


class ConfigPageFormTests(TestCase):
    """Unit tests for djblets.configforms.forms.ConfigPageForm."""

    def setUp(self):
        super(ConfigPageFormTests, self).setUp()

        request = RequestFactory().request()
        user = User.objects.create_user(username='test-user',
                                        password='test-user')
        page = TestPage(ConfigPagesView, request, user)

        self.form = TestForm(page, request, user)

    def test_initial_state(self):
        """Testing ConfigPageForm initial state"""
        self.assertEqual(self.form.fields['form_target'].initial, 'my-form')

    def test_set_initial(self):
        """Testing ConfigPageForm.set_initial"""
        self.form.set_initial({
            'field1': 'foo',
            'field2': 'bar',
        })

        self.assertEqual(self.form.fields['field1'].initial, 'foo')
        self.assertEqual(self.form.fields['field2'].initial, 'bar')

    def test_render(self):
        """Testing ConfigPageForm.render"""
        rendered = self.form.render()

        self.assertHTMLEqual(
            '<input id="id_form_target" name="form_target"'
            ' type="hidden" value="my-form">'
            '<div class="fields-row field-field1" id="row-field1">'
            ' <div class="field">'
            '  <label for="id_field1">Field 1:</label>'
            '  <input id="id_field1" name="field1" type="text">'
            ' </div>'
            '</div>'
            '<div class="fields-row field-field2" id="row-field2">'
            ' <div class="field">'
            '  <label for="id_field2">Field 2:</label>'
            '  <input id="id_field2" name="field2" type="text">'
            ' </div>'
            '</div>'
            '<input type="submit" class="btn" value="Save">',
            rendered)



"""Unit tests for djblets.extensions.views."""

from __future__ import unicode_literals

from django import forms
from django.conf.urls import url
from django.contrib import admin
from django.contrib.auth.models import User
from django.test.utils import override_settings
from kgb import SpyAgency

from djblets.extensions.extension import Extension
from djblets.extensions.forms import SettingsForm
from djblets.extensions.tests.base import ExtensionTestsMixin
from djblets.extensions.views import configure_extension
from djblets.testing.testcases import TestCase


class ViewTests(SpyAgency, ExtensionTestsMixin, TestCase):
    """Unit tests for djblets.extensions.views."""

    def setUp(self):
        class TestExtension(Extension):
            admin_urlconf = []
            is_configurable = True

        super(ViewTests, self).setUp()

        self.extension = self.setup_extension(TestExtension)

    def test_configure_extension_saving(self):
        """Testing configure_extension with saving settings"""
        class TestSettingsForm(SettingsForm):
            mykey = forms.CharField(max_length=100)

        self.spy_on(self.manager.get_enabled_extension,
                    call_fake=lambda *args: self.extension)

        urlpatterns[:] = [
            url('^config/$', configure_extension,
                {
                    'ext_class': type(self.extension),
                    'form_class': TestSettingsForm,
                    'extension_manager': self.manager,
                }),
            url('', admin.site.urls),
        ]

        User.objects.create_superuser(username='admin',
                                      password='admin',
                                      email='admin@example.com')

        with override_settings(ROOT_URLCONF=__name__):
            self.client.login(username='admin', password='admin')

            # Fetch once for the CSRF token.
            response = self.client.get('/config/')
            self.assertEqual(response.status_code, 200)

            # Save new settings in the form.
            response = self.client.post('/config/', data={
                'mykey': 'myvalue',
            })
            self.assertEqual(response.status_code, 302)

        self.assertEqual(self.extension.settings.get('mykey'), 'myvalue')


# A placeholder for any urlpatterns that need to be set for tests.
urlpatterns = []



from __future__ import unicode_literals

from django_evolution.mutations import RenameAppLabel


MUTATIONS = [
    RenameAppLabel('extensions', 'djblets_extensions',
                   legacy_app_label='extensions'),
]



"""Database models for integration configuration."""

from __future__ import unicode_literals

from django.db import models
from django.utils import timezone
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from djblets.db.fields import JSONField
from djblets.integrations.mixins import NeedsIntegrationManagerMixin


@python_2_unicode_compatible
class BaseIntegrationConfig(NeedsIntegrationManagerMixin, models.Model):
    """Base class for stored configuration for a particular integration.

    This contains configuration settings for a given instance of an
    integration, along with state indicating if that integration is to be
    enabled and user-specified identifying information.

    This is meant to be subclassed, in order to provide additional state
    that may be needed by an application.
    """

    integration_id = models.CharField(max_length=255, db_index=True)
    time_added = models.DateTimeField(default=timezone.now)
    last_updated = models.DateTimeField(default=timezone.now)
    name = models.CharField(max_length=255, blank=True, null=True)
    enabled = models.BooleanField(default=False, db_index=True)
    settings = JSONField(default={})
    extra_data = JSONField(default={})

    @property
    def integration(self):
        """The integration for the configuration."""
        return self.get_integration_manager().get_integration(
            self.integration_id)

    def get(self, key, default=None):
        """Return the setting for a given key.

        This will return the setting's stored value, or its default value for
        the integration if unset in these settings. If the setting isn't set
        in either place, the provided default will be returned instead.

        Args:
            key (unicode):
                The settings key.

            default (object):
                The default value, if not available elsewhere.

        Returns:
            The resulting value from settings, or the default value if not
            found.
        """
        try:
            return self.settings[key]
        except KeyError:
            return self.integration.default_settings.get(key, default)

    def set(self, key, value):
        """Set a value for the given settings key.

        This is equivalent to setting the value through the standard dictionary
        operators.

        Args:
            key (unicode):
                The key to set.

            value (object):
                The settings value.
        """
        self.settings[key] = value

    def __str__(self):
        """Return a string representation of this configuration.

        Returns:
            unicode:
            The string representation.
        """
        try:
            integration_name = self.integration.name
        except Exception:
            integration_name = self.integration_id

        return _('%(config_name)s (%(integration_name)s)') % {
            'config_name': self.name,
            'integration_name': integration_name,
        }

    class Meta:
        abstract = True
        verbose_name = _('Integration configuration')
        verbose_name_plural = _('Integration configurations')



from __future__ import unicode_literals

from django.conf.urls import include, url
from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import reverse
from django.utils import six
from django.utils.translation import ugettext as _

from djblets.extensions.errors import (DisablingExtensionError,
                                       EnablingExtensionError,
                                       InvalidExtensionError)
from djblets.extensions.models import RegisteredExtension
from djblets.extensions.signals import (extension_initialized,
                                        extension_uninitialized)
from djblets.urls.resolvers import DynamicURLResolver
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_permission_required,
                                       webapi_request_fields,
                                       webapi_response_errors)
from djblets.webapi.errors import (DOES_NOT_EXIST,
                                   ENABLE_EXTENSION_FAILED,
                                   DISABLE_EXTENSION_FAILED,
                                   PERMISSION_DENIED)
from djblets.webapi.fields import BooleanFieldType, StringFieldType
from djblets.webapi.resources.base import WebAPIResource


class ExtensionResource(WebAPIResource):
    """Provides information on installed extensions."""
    model = RegisteredExtension
    fields = {
        'author': {
            'type': StringFieldType,
            'description': 'The author of the extension.',
        },
        'author_url': {
            'type': StringFieldType,
            'description': "The author's website.",
        },
        'can_disable': {
            'type': BooleanFieldType,
            'description': 'Whether or not the extension can be disabled.',
        },
        'can_enable': {
            'type': BooleanFieldType,
            'description': 'Whether or not the extension can be enabled.',
        },
        'class_name': {
            'type': StringFieldType,
            'description': 'The class name for the extension.',
        },
        'enabled': {
            'type': BooleanFieldType,
            'description': 'Whether or not the extension is enabled.',
        },
        'installed': {
            'type': BooleanFieldType,
            'description': 'Whether or not the extension is installed.',
        },
        'loadable': {
            'type': BooleanFieldType,
            'description': 'Whether or not the extension is currently '
                           'loadable. An extension may be installed but '
                           'missing or may be broken due to a bug.',
        },
        'load_error': {
            'type': StringFieldType,
            'description': 'If the extension could not be loaded, this will '
                           'contain any errors captured while trying to load.',
        },
        'name': {
            'type': StringFieldType,
            'description': 'The name of the extension.',
        },
        'summary': {
            'type': StringFieldType,
            'description': "A summary of the extension's functionality.",
        },
        'version': {
            'type': StringFieldType,
            'description': 'The installed version of the extension.',
        },
    }
    name = 'extension'
    plural_name = 'extensions'
    uri_object_key = 'extension_name'
    uri_object_key_regex = r'[.A-Za-z0-9_-]+'
    model_object_key = 'class_name'

    allowed_methods = ('GET', 'PUT')

    def __init__(self, extension_manager):
        super(ExtensionResource, self).__init__()
        self._extension_manager = extension_manager
        self._dynamic_patterns = DynamicURLResolver()
        self._resource_url_patterns_map = {}

        # We want ExtensionResource to notice when extensions are
        # initialized or uninitialized, so connect some methods to
        # those signals.
        from djblets.extensions.signals import (extension_initialized,
                                                extension_uninitialized)
        extension_initialized.connect(self._on_extension_initialized)
        extension_uninitialized.connect(self._on_extension_uninitialized)

    def serialize_author_field(self, extension, *args, **kwargs):
        if extension.extension_class is None:
            return None

        return extension.extension_class.info.author

    def serialize_author_url_field(self, extension, *args, **kwargs):
        if extension.extension_class is None:
            return None

        return extension.extension_class.info.author_url

    def serialize_can_disable_field(self, extension, *args, **kwargs):
        return self._extension_manager.get_can_disable_extension(extension)

    def serialize_can_enable_field(self, extension, *args, **kwargs):
        return self._extension_manager.get_can_enable_extension(extension)

    def serialize_loadable_field(self, ext, *args, **kwargs):
        return (ext.extension_class is not None and
                ext.class_name not in self._extension_manager._load_errors)

    def serialize_load_error_field(self, extension, *args, **kwargs):
        s = self._extension_manager._load_errors.get(extension.class_name)

        if s:
            return s

        if extension.extension_class is None:
            return _(
                'This extension is not installed or could not be found. Try '
                're-installing it and then click "Scan for installed '
                'extensions."'
            )

        return None

    def serialize_name_field(self, extension, *args, **kwargs):
        if extension.extension_class is None:
            return extension.name
        else:
            return extension.extension_class.info.name

    def serialize_summary_field(self, extension, *args, **kwargs):
        if extension.extension_class is None:
            return None

        return extension.extension_class.info.summary

    def serialize_version_field(self, extension, *args, **kwargs):
        if extension.extension_class is None:
            return None

        return extension.extension_class.info.version

    @webapi_response_errors(DOES_NOT_EXIST, PERMISSION_DENIED)
    @webapi_login_required
    def get_list(self, request, *args, **kwargs):
        """Returns the list of known extensions.

        Each extension in the list has been installed, but may not be
        enabled.
        """
        return WebAPIResource.get_list(self, request, *args, **kwargs)

    def get_links(self, resources=[], obj=None, request=None, *args, **kwargs):
        links = super(ExtensionResource, self).get_links(
            resources, obj, request=request, *args, **kwargs)

        if request and obj:
            admin_base_href = '%s%s' % (
                request.build_absolute_uri(reverse('extension-list')),
                obj.class_name)

            extension_cls = obj.extension_class

            if extension_cls:
                extension_info = extension_cls.info

                if extension_info.is_configurable:
                    links['admin-configure'] = {
                        'method': 'GET',
                        'href': '%s/config/' % admin_base_href,
                    }

                if extension_info.has_admin_site:
                    links['admin-database'] = {
                        'method': 'GET',
                        'href': '%s/db/' % admin_base_href,
                    }

        return links

    @webapi_login_required
    @webapi_permission_required('extensions.change_registeredextension')
    @webapi_response_errors(PERMISSION_DENIED, DOES_NOT_EXIST,
                            ENABLE_EXTENSION_FAILED, DISABLE_EXTENSION_FAILED)
    @webapi_request_fields(
        required={
            'enabled': {
                'type': BooleanFieldType,
                'description': 'Whether or not to make the extension active.'
            },
        },
    )
    def update(self, request, *args, **kwargs):
        """Updates the state of the extension.

        If ``enabled`` is true, then the extension will be enabled, if it is
        not already. If false, it will be disabled.
        """
        # Try to find the registered extension
        try:
            registered_extension = self.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        extension_id = registered_extension.class_name

        if kwargs.get('enabled'):
            try:
                self._extension_manager.enable_extension(extension_id)
            except EnablingExtensionError as e:
                err = ENABLE_EXTENSION_FAILED.with_message(six.text_type(e))

                return err, {
                    'load_error': e.load_error,
                    'needs_reload': e.needs_reload,
                }
            except InvalidExtensionError as e:
                return ENABLE_EXTENSION_FAILED.with_message(six.text_type(e))
        else:
            try:
                self._extension_manager.disable_extension(extension_id)
            except (DisablingExtensionError, InvalidExtensionError) as e:
                return DISABLE_EXTENSION_FAILED.with_message(six.text_type(e))

        # Refetch extension, since the ExtensionManager may have changed
        # the model.
        registered_extension = \
            RegisteredExtension.objects.get(pk=registered_extension.pk)

        return 200, {
            self.item_result_key: registered_extension
        }

    def get_url_patterns(self):
        # We want extension resource URLs to be dynamically modifiable,
        # so we override get_url_patterns in order to capture and store
        # a reference to the url_patterns at /api/extensions/.
        url_patterns = super(ExtensionResource, self).get_url_patterns()
        url_patterns += [self._dynamic_patterns]

        return url_patterns

    def get_related_links(self, obj=None, request=None, *args, **kwargs):
        """Returns links to the resources provided by the extension.

        The result should be a dictionary of link names to a dictionary of
        information. The information should contain:

        * 'method' - The HTTP method
        * 'href' - The URL
        * 'title' - The title of the link (optional)
        * 'resource' - The WebAPIResource instance
        * 'list-resource' - True if this links to a list resource (optional)
        """
        links = {}

        if obj and obj.enabled:
            extension = obj.get_extension_class()

            if not extension:
                return links

            for resource in extension.resources:
                links[resource.name_plural] = {
                    'method': 'GET',
                    'href': "%s%s/" % (
                        self.get_href(obj, request, *args, **kwargs),
                        resource.uri_name),
                    'resource': resource,
                    'list-resource': not resource.singleton,
                }

        return links

    def _attach_extension_resources(self, extension):
        """
        Attaches an extension's resources to /api/extensions/{extension.id}/.
        """

        # Bail out if there are no resources to attach
        if not extension.resources:
            return

        if extension in self._resource_url_patterns_map:
            # This extension already had its urlpatterns
            # mapped and attached.  Nothing to do here.
            return

        # We're going to store references to the URL patterns
        # that are generated for this extension's resources.
        self._resource_url_patterns_map[extension] = []

        # For each resource, generate the URLs
        for resource in extension.resources:
            self._resource_url_patterns_map[extension].extend([
                url(r'^%s/%s/' % (extension.id, resource.uri_name),
                    include(resource.get_url_patterns())),
            ])

        self._dynamic_patterns.add_patterns(
            self._resource_url_patterns_map[extension])

    def _unattach_extension_resources(self, extension):
        """
        Unattaches an extension's resources from
        /api/extensions/{extension.id}/.
        """

        # Bail out if there are no resources for this extension
        if not extension.resources:
            return

        # If this extension has never had its resource URLs
        # generated, then we don't have anything to worry
        # about.
        if extension not in self._resource_url_patterns_map:
            return

        # Remove the URL patterns
        self._dynamic_patterns.remove_patterns(
            self._resource_url_patterns_map[extension])

        # Delete the URL patterns so that we can regenerate
        # them when the extension is re-enabled.  This is to
        # avoid caching incorrect URL patterns during extension
        # development, when extension resources are likely to
        # change.
        del self._resource_url_patterns_map[extension]

    def _on_extension_initialized(self, sender, ext_class=None, **kwargs):
        """
        Signal handler that notices when an extension has been initialized.
        """
        self._attach_extension_resources(ext_class)

    def _on_extension_uninitialized(self, sender, ext_class=None, **kwargs):
        """
        Signal handler that notices and reacts when an extension
        has been uninitialized.
        """
        self._unattach_extension_resources(ext_class)


class ExtensionRootResourceMixin(object):
    """Mixin for Root Resources making use of Extension Resources.

    As extensions are able to provide their own API resources, this mixin
    allows a root resource to generate URI templates for non built-in
    resources.

    See Also:
        :py:class:`~djblets.webapi.resources.root.RootResource`
    """

    def __init__(self, *args, **kwargs):
        """Initialize the extension resource mixin to listen for changes.

        Args:
            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.
        """
        super(ExtensionRootResourceMixin, self).__init__(*args, **kwargs)

        extension_initialized.connect(
            self._generate_extension_uris_for_template)
        extension_uninitialized.connect(
            self._remove_extension_uris_from_template)

    def get_extension_resource(self):
        """Return the associated extension resource.

        Subclasses using this mixin must implement this method.

        Returns:
            djblets.extensions.resources.ExtensionResource:
            The extension resource associated with the root resource.
        """
        raise NotImplementedError

    def _generate_extension_uris_for_template(self, ext_class, **kwargs):
        """Generate URI templates for a newly enabled extension.

        Args:
            ext_class djblets.extensions.extension.Extension:
                The extension being added to the URI templates.

            **kwargs (dict):
                Additional keyword arguments.
        """
        ext_resource = self.get_extension_resource()

        for resource in ext_class.resources:
            partial_href = '%s/%s/' % (ext_class.id, resource.uri_name)

            for entry in self.walk_resources(resource, partial_href):
                self.register_uri_template(entry.name, entry.list_href,
                                           ext_resource)

    def _remove_extension_uris_from_template(self, ext_class, **kwargs):
        """Remove the URI templates of an extension when disabled.

        Args:
            ext_class djblets.extensions.extension.Extension:
                The extension being removed from the URI templates.

            **kwargs (dict):
                Additional keyword arguments.
        """
        ext_resource = self.get_extension_resource()

        for resource in ext_class.resources:
            partial_href = '%s/%s/' % (ext_class.id, resource.uri_name)

            for entry in self.walk_resources(resource, partial_href):
                self.unregister_uri_template(entry.name, ext_resource)



"""Logging support.

This can be used to offer some advanced logging capabilities.

Note that this predates Django's modern logging support, and is here primarily
for compatibility.


Settings
========

The following settings control logging.


LOGGING_ENABLED
---------------

Default: ``False``

Sets whether or not logging is enabled.


LOGGING_DIRECTORY
-----------------

Default: ``None``

Specifies the directory that log files should be stored in.
This directory must be writable by the process running Django.


LOGGING_NAME
------------

Default: ``None``

The name of the log files, excluding the extension and path. This will
usually be the name of the website or web application. The file extension
will be automatically appended when the file is written.


LOGGING_ALLOW_PROFILING
-----------------------

Default: ``False``

Specifies whether or not code profiling is allowed. If True, visiting
any page with a ``?profiling=1`` parameter in the URL will cause the
request to be profiled and stored in a ``.prof`` file using the defined
``LOGGING_DIRECTORY`` and ``LOGGING_NAME``.


LOGGING_LINE_FORMAT
-------------------

Default: ``"%(asctime)s - %(levelname)s - %(message)s"``

The format for lines in the log file. See Python's :py:mod:`logging`
documentation for possible values in the format string.


LOGGING_PAGE_TIMES
------------------

Default: ``False``

If enabled, page access times will be logged. Specifically, it will log
the initial request, the finished render and response, and the total
time it look.

The username and page URL will be included in the logs.


LOGGING_LEVEL
-------------

Default: ``"DEBUG"``

The minimum level to log. Possible values are ``"DEBUG"``, ``"INFO"``,
``"WARNING"``, ``"ERROR"`` and ``"CRITICAL"``.


LOGGING_BLACKLIST
-----------------

Default: ``['django.db.backends']``

A list of logger names to exclude from the logs. Each logger with the given
name will be filtered out, along with any descendents of those loggers.
"""

from __future__ import unicode_literals

import inspect
import logging
import logging.handlers
import os
import sys
from datetime import datetime
from functools import update_wrapper

from django.conf import settings


default_app_config = 'djblets.log.apps.LogAppConfig'


_logging_setup = False
_profile_log = None

DEFAULT_LOG_LEVEL = "DEBUG"
DEFAULT_LINE_FORMAT = \
    "%(asctime)s - %(levelname)s - %(request_info)s - %(name)s - %(message)s"
DEFAULT_REQUEST_FORMAT = '%(user)s - %(path)s'


class TimedLogInfo(object):
    """
    A utility class created by ``log_timed`` that handles the timed logging
    functionality and provides a way to end the timed logging operation.
    """
    def __init__(self, message, warning_at, critical_at, default_level,
                 log_beginning, request):
        self.message = message
        self.warning_at = warning_at
        self.critical_at = critical_at
        self.default_level = default_level
        self.start_time = datetime.utcnow()
        self.request = request

        if log_beginning:
            logging.log(self.default_level, "Begin: %s" % self.message,
                        request=self.request)

    def done(self):
        """
        Stops the timed logging operation. The resulting time of the
        operation will be written to the log file. The log level depends
        on how long the operation takes.
        """
        delta = datetime.utcnow() - self.start_time
        level = self.default_level

        if delta.seconds >= self.critical_at:
            level = logging.CRITICAL
        elif delta.seconds >= self.warning_at:
            level = logging.WARNING

        logging.log(self.default_level, "End: %s" % self.message,
                    request=self.request)
        logging.log(level, '%s took %d.%06d seconds' % (self.message,
                                                        delta.seconds,
                                                        delta.microseconds),
                    request=self.request)


class RequestLogFormatter(logging.Formatter):
    def __init__(self, request_fmt, *args, **kwargs):
        super(RequestLogFormatter, self).__init__(*args, **kwargs)
        self.request_fmt = request_fmt

    def format(self, record):
        record.request_info = self.format_request(
            getattr(record, 'request', None))

        return logging.Formatter.format(self, record)

    def format_request(self, request):
        """Return formatted request information for the log message.

        The returned string will be based off the configured
        ``settings.LOGGING_REQUEST_FORMAT`` and keys in the request, if
        found.

        If anything from the request is missing, an empty string will be
        returned.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            unicode:
            The request-specific string to include in the log message. This
            may be empty.
        """
        s = ''

        if request:
            try:
                s = self.request_fmt % request.__dict__
            except KeyError:
                # The request isn't populated with the keys expected in the
                # format string. Assume that we're logging before some
                # middleware or app configuration has had a chance to set
                # things up, and return an empty string.
                pass

        return s


class BlacklistFilter(logging.Filter):
    """Blacklists the provided loggers (and their children) from logging."""

    def __init__(self, names):
        """Initialize the filter.

        Args:
            names (list of unicode):
                A list of logger names. Each logger (and their children) will
                be excluded from the logs.
        """
        self._filters = [
            logging.Filter(name)
            for name in names
        ]

    def filter(self, record):
        """Return whether this record should be logged.

        The record is only logged if it's not in the list of any of the
        loggers on the blacklist, and if it doesn't have a parent logger
        listed.

        Args:
            record (logging.LogRecord):
                The record to filter.

        Returns:
            bool:
            ``True`` if the record can be logged. ``False`` if it must be
            ignored.
        """
        return all(not log_filter.filter(record)
                   for log_filter in self._filters)


def init_logging():
    """
    Sets up the main loggers, if they haven't already been set up.
    """
    global _logging_setup

    if _logging_setup:
        return

    enabled = getattr(settings, 'LOGGING_ENABLED', False)
    log_directory = getattr(settings, 'LOGGING_DIRECTORY', None)
    log_name = getattr(settings, 'LOGGING_NAME', None)

    if not enabled or not log_directory or not log_name:
        return

    log_level_name = getattr(settings, 'LOGGING_LEVEL',
                             DEFAULT_LOG_LEVEL)
    log_level = logging.getLevelName(log_level_name)
    request_format_str = getattr(settings, 'LOGGING_REQUEST_FORMAT',
                                 DEFAULT_REQUEST_FORMAT)
    format_str = getattr(settings, 'LOGGING_LINE_FORMAT',
                         DEFAULT_LINE_FORMAT)
    log_blacklist = getattr(settings, 'LOGGING_BLACKLIST', [
        'django.db.backends',
    ])

    log_path = os.path.join(log_directory, log_name + ".log")

    formatter = RequestLogFormatter(request_format_str, format_str)
    logging_to_stderr = False

    root = logging.getLogger()

    if log_path:
        try:
            if sys.platform == 'win32':
                handler = logging.FileHandler(log_path)
            else:
                handler = logging.handlers.WatchedFileHandler(log_path)

            logging_to_stderr = False
        except IOError:
            handler = logging.StreamHandler()
            logging_to_stderr = True

        handler.setLevel(log_level)
        handler.setFormatter(formatter)

        root.addHandler(handler)
        root.setLevel(log_level)

    if logging_to_stderr:
        logging.warning("Could not open logfile %s. Logging to stderr",
                        log_path)

    if settings.DEBUG and not logging_to_stderr:
        # In DEBUG mode, log to the console as well.
        console_log = logging.StreamHandler()
        console_log.setLevel(log_level)
        console_log.setFormatter(formatter)
        root.addHandler(console_log)

        logging.debug("Logging to %s with a minimum level of %s",
                      log_path, log_level_name)

    for handler in root.handlers:
        handler.addFilter(BlacklistFilter(log_blacklist))

    _logging_setup = True


def init_profile_logger():
    """
    Sets up the profiling logger, if it hasn't already been set up.
    """
    global _profile_log

    enabled = getattr(settings, 'LOGGING_ENABLED', False)
    log_directory = getattr(settings, 'LOGGING_DIRECTORY', None)
    log_name = getattr(settings, 'LOGGING_NAME', None)

    if (enabled and log_directory and log_name and not _profile_log and
        getattr(settings, "LOGGING_ALLOW_PROFILING", False)):

        filename = os.path.join(log_directory, log_name + ".prof")

        if sys.platform == 'win32':
            handler = logging.FileHandler(filename)
        else:
            handler = logging.handlers.WatchedFileHandler(filename)

        handler.setLevel(logging.INFO)
        handler.setFormatter(logging.Formatter("%(asctime)s %(message)s"))

        _profile_log = logging.getLogger("profile")
        _profile_log.addHandler(handler)


def restart_logging():
    """
    Restarts the logging. The next page view will set up the loggers
    based on any new settings.
    """
    global _logging_setup

    logging.debug('Reloading logging settings')

    for logger_id in ('profile', ''):
        logger = logging.getLogger(logger_id)

        while logger.handlers:
            handler = logger.handlers[0]
            handler.flush()
            logger.removeHandler(handler)

    _logging_setup = False

    init_logging()


def log_timed(message, warning_at=5, critical_at=15,
              log_beginning=True, default_level=logging.DEBUG,
              request=None):
    """
    Times an operation, displaying a log message before and after the
    operation. The log level for the final log message containing the
    operation runtime will be based on the runtime, the ``warning_at`` and
    the ``critical_at`` parameters.
    """
    return TimedLogInfo(message, warning_at, critical_at, default_level,
                        log_beginning, request)


# Add support in Logger to take a request= parameter. This will apply to the
# root logger and all instances/subclasses of logging.Logger.
#
# Ideally, we would use logging.setLoggerClass() to set a custom Logger that
# understands request=, but the root logger wouldn't be affected and anything
# that called logging.getLogger('...') before this file was imported wouldn't
# benefit.
def _Logger_log(self, *args, **kwargs):
    extra = kwargs.pop('extra', {})
    request = kwargs.pop('request', None)

    if request:
        extra['request'] = request

    kwargs['extra'] = extra

    _old_log(self, *args, **kwargs)

_old_log = logging.Logger._log

if _old_log is not _Logger_log:
    update_wrapper(_Logger_log, logging.Logger._log)
    logging.Logger._log = _Logger_log


# On some versions of Python (2.6.x, 2.7.0-2.7.5, 3.0.x, and 3.1.x),
# Logger.exception/logging.exception doesn't support keyword arguments,
# which impacts not only request= but also extra=. We need to patch these.
def _has_keywords(func):
    if hasattr(inspect, 'getfullargspec'):
        argspec = inspect.getfullargspec(func)

        return (argspec.varkw is not None or
                argspec.kwonlyargs is not None)
    else:
        return inspect.getargspec(func).keywords is not None

if not _has_keywords(logging.exception):
    def _logging_exception(msg, *args, **kwargs):
        kwargs['exc_info'] = 1
        logging.error(msg, *args, **kwargs)

    update_wrapper(_logging_exception, logging.exception)
    logging.exception = _logging_exception

if not _has_keywords(logging.Logger.exception):
    def _Logger_exception(self, msg, *args, **kwargs):
        kwargs['exc_info'] = 1
        self.error(msg, *args, **kwargs)

    update_wrapper(_Logger_exception, logging.Logger.exception)
    logging.Logger.exception = _Logger_exception



from __future__ import unicode_literals

from django.utils import six

from djblets.integrations.integration import Integration
from djblets.integrations.manager import IntegrationManager
from djblets.integrations.tests.models import IntegrationConfig
from djblets.integrations.tests.testcases import IntegrationsTestCase


class DummyIntegration1(Integration):
    name = 'Dummy'

    default_settings = {
        'foo': 'default-foo',
    }

    def initialize(self):
        pass


class IntegrationConfigTests(IntegrationsTestCase):
    """Unit tests for djblets.integrations.models.BaseIntegrationConfig."""

    def setUp(self):
        super(IntegrationConfigTests, self).setUp()

        self.manager = IntegrationManager(IntegrationConfig)
        self.integration = \
            self.manager.register_integration_class(DummyIntegration1)
        self.config = self.integration.create_config(name='Test Config')
        self.config.manager = self.manager

    def test_integration(self):
        """Testing BaseIntegrationConfig.integration"""
        self.assertEqual(self.config.integration, self.integration)

    def test_get(self):
        """Testing BaseIntegrationConfig.get"""
        self.config.settings['foo'] = 'bar'
        self.assertEqual(self.config.get('foo'), 'bar')

    def test_get_with_integration_defaults(self):
        """Testing BaseIntegrationConfig.get with integration defaults"""
        self.assertEqual(self.config.get('foo'), 'default-foo')

    def test_get_with_default(self):
        """Testing BaseIntegrationConfig.get with default"""
        self.assertEqual(self.config.get('new-key', 'my-default'),
                         'my-default')

    def test_set(self):
        """Testing BaseIntegrationConfig.set"""
        self.config.set('my-key', 'my-value')
        self.assertEqual(self.config.settings.get('my-key'), 'my-value')

    def test_str_with_available_integration(self):
        """Testing BaseIntegrationConfig.__str__ with available integration"""
        self.assertEqual(six.text_type(self.config), 'Test Config (Dummy)')

    def test_str_without_available_integration(self):
        """Testing BaseIntegrationConfig.__str__ without available integration
        """
        self.config.integration_id = 'foo.bar.BadIntegration'

        self.assertEqual(six.text_type(self.config),
                         'Test Config (foo.bar.BadIntegration)')



"""An avatar service used as a fallback to show something basic for a user."""

from __future__ import division, unicode_literals

from django.template.loader import render_to_string
from django.utils.translation import ugettext_lazy as _

from djblets.avatars.services.base import AvatarService


class FallbackService(AvatarService):
    """An avatar service used as a fallback.

    This will display a simple avatar showing the first two characters of the
    user's username, displayed on top of a background with a color based on the
    username.

    This will automatically be used as a fallback if no other avatar backends
    are usable for the user.
    """

    avatar_service_id = 'fallback'
    name = _('Fallback')
    template_name = 'avatars/fallback.html'
    hidden = True

    def render(self, request, user, size, template_name=None):
        """Render a user's avatar to HTML.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            user (django.contrib.auth.models.User):
                The user for whom the avatar is to be rendered.

            size (int):
                The requested avatar size (height and width) in pixels.

            template_name (unicode, optional):
                The name of the template to use for rendering.

        Returns:
            django.utils.safestring.SafeText:
            The rendered avatar HTML.
        """
        return render_to_string(template_name or self.template_name, {
            'bg': self.get_bg_color(user),
            'font_size': '%spx' % (size // 3),
            'size': size,
            'text': user.username[:2].upper(),
            'user': user,
        })

    def get_bg_color(self, user):
        """Return a background color for the avatar.

        This will compute a basic HSL color for the avatar, based on the
        username.

        Args:
            user (django.contrib.auth.models.User):
                The user to generate the color for.

        Returns:
            unicode:
            The resulting HSL color definition.
        """
        i = 0

        for c in user.username:
            i = ord(c) + ((i << 5) - i)

        i = i % 360

        return 'hsl(%s, 60%%, 80%%)' % i

    def get_avatar_urls_uncached(self, user, size):
        """Return the Gravatar URLs for the requested user.

        Args:
            user (django.contrib.auth.models.User):
                The user whose avatar URLs are to be fetched.

            size (int):
                The size (in pixels) the avatar is to be rendered at.

        Returns
            dict:
            A dictionary containing the URLs of the user's avatars at normal-
            and high-DPI.
        """
        return {
            '1x': '',
            '2x': '',
            '3x': '',
        }

    def get_etag_data(self, user):
        """Return the ETag data for the user's avatar.

        Args:
            user (django.contrib.auth.models.User):
                The user.

        Returns:
            list of unicode:
            The uniquely identifying information for the user's avatar.
        """
        return [self.avatar_service_id, user.username]
